/* bsocket_l.C
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: bsocket_l.C,v 1.2 2003/10/02 19:29:56 cos Exp $
 *
 */
/******************************************************************
 *               bSocket Class - Light version                    *
 *****************************************************************/

// by Bluffy
#include <unistd.h>
#include <memory.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <pthread.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include "bsocket_l.h"

// for loops: set Online to 'stop' and tries 10 times (stop_wait interval)
// to look whether set to 0
#define STOP_WAIT       1
#define STAT_CONNECTED  1
#define STAT_WAITLOOP   2
#define STAT_RECVLOOP   3
#define STAT_STOP       -1
#define BUFFERLENGTH    10000
#define NULL 0

// Constructor
bSocket::bSocket(bool udp)
{
  UDP=udp;
  Online=false;
  InternalSocket=NULL;
}


bSocket::~bSocket()
{
  if (InternalSocket) close(InternalSocket);
}


bool bSocket::Disconnect()
{
  if (Online==STAT_WAITLOOP) return StopWaitLoop();
  if (Online==STAT_RECVLOOP) return StopReceiveLoop();
  if (InternalSocket) close(InternalSocket);
  InternalSocket=NULL;
  Online=0;
  return true;
}


// Connect the socket to the IRC server
bool bSocket::Connect(char *HostName,int Port,char * interface)
{
  if (InternalSocket) close(InternalSocket);
  if (!UDP) InternalSocket=socket(AF_INET,SOCK_STREAM,IPPROTO_IP);
  else	InternalSocket=socket(AF_INET,SOCK_DGRAM,IPPROTO_IP);
  sockaddr_in SockAddr;
  memset(&SockAddr,0,sizeof(sockaddr_in));
  SockAddr.sin_family = AF_INET;
  SockAddr.sin_port= htons(Port);
  if (interface)
    {
      sockaddr_in sa;
      sa.sin_family=AF_INET;
      sa.sin_port=0;//on s'en fou
      sa.sin_addr.s_addr=inet_addr(interface);
      bind(InternalSocket,(sockaddr *)&sa,sizeof(sockaddr_in));
    }
  hostent *hp=gethostbyname(HostName);
  if (!hp)
    {
      SockAddr.sin_addr.s_addr=inet_addr(HostName);
      if (SockAddr.sin_addr.s_addr==INADDR_NONE)
	return false;
    }  
  else memcpy(&SockAddr.sin_addr,hp->h_addr,4);  
  if (connect(InternalSocket,(sockaddr *)&SockAddr,sizeof(sockaddr_in))==-1)
    {
      int l=errno;
      return false;
    }
  return true;
}


unsigned long bSocket::Receive(char *Buffer,unsigned long MaxLen)
{
  if (!InternalSocket)
    return 0;
  unsigned long Result=recv(InternalSocket,Buffer,MaxLen,0);
  return Result;
}


unsigned long bSocket::SlowSend(const char * Buffer,unsigned long len)
{
  if (!this) return 0;
  for (int i=0;i<len;i++) 
    {
      if (send(InternalSocket,&Buffer[i],1,0)!=1) return 0;
      usleep(20000);
    }
  return 1;
}


bool bSocket::InternalSend(const char *Buffer, unsigned long Len)
{
  long Sent;
  if (InternalSocket==NULL)
    return false;
  Sent=send(InternalSocket,Buffer,Len,0);
  if (Sent!=Len)
    return false;
  return true;
}


unsigned long bSocket::Send(const char *ToSend)
{
  if (!this) return 0;
  return InternalSend(ToSend,strlen(ToSend));
}


unsigned long bSocket::Send(const char *ToSend,unsigned long Len)
{
  if (!this) return 0;
  return InternalSend(ToSend,Len);
}


void * clt(void * param)
{
  ((bSocket*)param)->WaitLoop();
  return 0;
}


void * clt2(void * param)
{
  ((bSocket*)param)->ReceiveLoop();
  return 0;
}


bool bSocket::CreateWaitLoop(int Port,tOnConnect OnConnect,tOnError onerror)
{
  OnError=onerror;
  if (Online) Disconnect();
  if (InternalSocket) close(InternalSocket);
  //on bind
  if (!UDP) InternalSocket=socket(AF_INET,SOCK_STREAM,IPPROTO_IP);
  else	InternalSocket=socket(AF_INET,SOCK_DGRAM,IPPROTO_IP);
  if (InternalSocket==-1) return false;
  sockaddr_in localsin;
  localsin.sin_family=AF_INET;
  localsin.sin_addr.s_addr=INADDR_ANY;
  localsin.sin_port=htons(Port);
  if (bind(InternalSocket,(sockaddr *)&localsin,sizeof(sockaddr_in))==-1) return false;
  listen(InternalSocket,SOMAXCONN);
  pthread_t inu;
  Online=STAT_WAITLOOP;
  CallBack=(void *)OnConnect;
  pthread_create(&inu,NULL,&clt,this);
  return true;
}


bool bSocket::Bind(int Port,char * interface)
{
  if (Online) Disconnect();
  if (InternalSocket) close(InternalSocket);
  if (!UDP) InternalSocket=socket(AF_INET,SOCK_STREAM,IPPROTO_IP);
  else	InternalSocket=socket(AF_INET,SOCK_DGRAM,IPPROTO_IP);
  if (InternalSocket==-1) return false;
  sockaddr_in localsin;
  localsin.sin_family=AF_INET;
  localsin.sin_addr.s_addr=interface?inet_addr(interface):INADDR_ANY;
  localsin.sin_port=htons(Port);
  if (bind(InternalSocket,(sockaddr *)&localsin,sizeof(sockaddr_in))==-1) return false;
  return true;
}


bool bSocket::StopWaitLoop()
{
  if (Online!=STAT_WAITLOOP) return false;
  Online=STAT_STOP;
  sleep(0);
  for (int i=10;i&&Online;i--) sleep(STOP_WAIT);
  return !Online;
}


bool bSocket::StopReceiveLoop()
{
  if (Online!=STAT_RECVLOOP) return false;
  Online=STAT_STOP;
  sleep(0);
  for (int i=10;i&&Online;i--) sleep(STOP_WAIT);
  return !Online;
}


bool bSocket::CreateReceiveLoop(tOnReceive OnReceive,tOnError onerror)
{
  OnError=onerror;
  // DO NOT DECONNECT OF COURSE if (Online) Disconnect();
  Online=STAT_RECVLOOP;
  CallBack=(void *)OnReceive;
  pthread_t inu;
  pthread_create(&inu,0,&clt2,this);
  return true;
}

void bSocket::ReceiveLoop()
{
  void * buff=malloc(BUFFERLENGTH);
  unsigned long len;
  while (Online==STAT_RECVLOOP)
    {
      len=Receive((char *)buff,BUFFERLENGTH);
      if (len==-1) break;
      if (!len) break; // socket closed gracefuly
      ((tOnReceive)CallBack)(this,buff,len);
    }
  if (Online==STAT_RECVLOOP)
    {
      unsigned long err=len?errno:0;
      if (OnError) OnError(this,err);
    }
  if (InternalSocket) close(InternalSocket);
  InternalSocket=NULL;
  Online=0;
}


void bSocket::WaitLoop()
{
  while (Online==STAT_WAITLOOP)
    {
      sockaddr_in sin;
      int sz=sizeof(sockaddr_in);
      memset(&sin,0,sz);
      int  s=accept(InternalSocket,(sockaddr*)&sin,(socklen_t*)&sz);
      // SOCKET s=accept(InternalSocket,NULL,0);
      if (s==-1)
	break;
      bSocket * nb=new bSocket(UDP);
      nb->Online=STAT_CONNECTED;
      memcpy(&nb->Destination,&sin,sizeof(sockaddr_in));
      nb->InternalSocket=s;
      ((tOnConnect)CallBack)(this,nb);
    }
  if ((Online==STAT_WAITLOOP)&&(OnError)) OnError(this,errno);
  if (InternalSocket) close(InternalSocket);
  InternalSocket=NULL;
  Online=0;
}


void bSocket::GetLocalIp(char *ip)
{
  sockaddr_in addr;
  int sz=sizeof(sockaddr_in);
  getsockname(InternalSocket,(sockaddr *)&addr,(socklen_t *)&sz);
  strcpy(ip,inet_ntoa(addr.sin_addr));
}


bool bSocket::GetDestination(char * Buffer)
{
  if (Online!=STAT_CONNECTED) return false;
  strcpy(Buffer,inet_ntoa((in_addr)Destination.sin_addr));
  return true;
}
