/* dnsquery.C
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: dnsquery.C,v 1.5 2006/04/10 21:04:16 cvs Exp $
 *
 */

#include <stdio.h>
#include <alloc.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "bsocket_l.h"
#include "dnsquery.h"
#include "net.h"
#define DEBUG 0

struct QueryTail
{
  unsigned short QType;
  unsigned short QClass;
};

#pragma allign 8


struct TFlags
{
  unsigned  RD:1; // recursion waznted
  unsigned  RA:1; // recursion availables
  unsigned  Z:3; // reserved
  unsigned  RCode:4; // responce code
  unsigned  Response:1;
  unsigned  OpCode:4;
  unsigned  AA:1;
  unsigned  Truncated:1;
};


struct TDNS
{
  unsigned short TransactionID;
  unsigned short Flags;
  unsigned short Question;
  unsigned short AnswerRR;
  unsigned short AuthorityRR;
  unsigned short AdditionalRR;
};


#define IPEXTRACT(a) (a[0]<<24)+(a[1]<<16)+(a[2]<<8)+a[3]


/*
 * Functions
 */



#define TYPE_PTR        12
#define QTYPE_AXFR      252
#define TYPE_A          1
#define TYPE_NS         2
#define TYPE_CNAME      5
#define TYPE_MX         15
#define CLASS_IN        1
#define OP_QUERY        0
#define OP_RQUERY       1
#define OP_STATUS       2

char * DNSCanIpInsert(char * src,unsigned long val);


unsigned long RDNSQuery(unsigned long IP,unsigned long server,char * Buffer)
{
  static unsigned short TransactionId=0x0002;
  static bSocket * bs=NULL;
  if (!bs) bs=new bSocket(true);
  // Protection multithread
  sleep(0);
  unsigned short CId=++TransactionId; 
  // creation requette
  TDNS *Query=(TDNS *)malloc(sizeof(TDNS)+50);
  memset(Query,0,sizeof(TDNS)+30);
  Query->TransactionID=CId;
  TFlags * Flags=(TFlags *)&Query->Flags;
  Flags->OpCode=OP_QUERY;
  Flags->RD=1;
  Query->Question=0x0100;
  // on colle la vraie query
  char * qname=&(((char *) Query)[sizeof(TDNS)]);
  qname=DNSCanIpInsert(qname,IP);
  QueryTail * tl=(QueryTail *)qname;
  tl->QClass=CLASS_IN<<8;
  tl->QType=TYPE_PTR<<8;
  unsigned long length=(unsigned long)tl-(unsigned long)Query+sizeof(QueryTail);
  // requette prete: envoi
	
  bool ok=bs->Connect(IPToStr(server),53);
  if (!ok) return ERROR_SOCK_CONN;
  ok=bs->Send((char *)Query,length);
  if (!ok) return ERROR_SOCK_SEND;
  TDNS * Re=(TDNS *)malloc(513);
  int retcode=bs->Receive((char *)Re,513);
  if (retcode==-1) return ERROR_SOCK_RECV;
  bs->Disconnect();
  // plus qu'a regarder le resultat
  unsigned long RCode=((char *)Re)[3]&0xf;
  if (RCode) return ERROR_DNS;
  return ERROR_OK;
}


char * DNSCanIpInsert(char * src,unsigned long ip)
{
  static char * IPCAN1="IN-ADDR";
  static char * IPCAN2="ARPA";
  char tbuff[100];
  sprintf(tbuff,"%u",ip&0xff);
  *src=strlen(tbuff);
  src++;
  strcpy(src,tbuff);
  src+=strlen(tbuff);
  sprintf(tbuff,"%u",(ip>>8)&0xff);
  *src=strlen(tbuff);
  src++;
  strcpy(src,tbuff);
  src+=strlen(tbuff);
  sprintf(tbuff,"%u",(ip>>16)&0xff);
  *src=strlen(tbuff);
  src++;
  strcpy(src,tbuff);
  src+=strlen(tbuff);
  sprintf(tbuff,"%u",(ip>>24)&0xff);
  *src=strlen(tbuff);
  src++;
  strcpy(src,tbuff);
  src+=strlen(tbuff);
  *src=strlen(IPCAN1);
  src++;
  strcpy(src,IPCAN1);
  src+=strlen(IPCAN1);
  *src=strlen(IPCAN2);
  src++;
  strcpy(src,IPCAN2);
  src+=strlen(IPCAN2);
  *src=0;
  src++;
  return src;
}
