/* lc.C
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: lc.C,v 1.2 2003/10/02 19:29:56 cos Exp $
 *
 */
/******************************************************************
 *            LC - Petite bibliotheque de fonctions               *
 *                    pour les listes chainee                     *
 *****************************************************************/

// By Bluffy
#ifndef _WIN32 
#define _WIN32
#endif
#include "lc.h"
#include "lc.hpp"

// Initialise l'objet liste chaine
void ListeChainee::InitLC(void)
{
  if (!this) return;
  SetPrev(NULL);
  SetNext(NULL);
}

// Destructeur
ListeChainee::~ListeChainee()
{
  EradicateFromLC();
}

// fonctions triviales
ListeChainee *ListeChainee::GetPrev()
{
  if (!this) return NULL;
  try
    {return Prev;}
  catch(...){return NULL;}
}


ListeChainee *ListeChainee::GetNext()
{
  if (!this) return NULL;
  try
    {return Next;}
  catch(...){return NULL;}
}


void ListeChainee::SetPrev(ListeChainee *prior)
{       
  if (!this) return;
  Prev=prior;
}


void ListeChainee::SetNext(ListeChainee *next)
{
  if (!this) return;
  Next=next;
}

// renvoit le premier lment de la liste chaine
ListeChainee *ListeChainee::GetFirst()
{
  if (!this) return NULL;
  ListeChainee *result=this;
  while (true)
    {
      if (!result->GetPrev()) return result;
      result=result->GetPrev();
    }
}

// renvoit le dernier
ListeChainee *ListeChainee::GetLast()
{
  if (!this) return NULL;
  ListeChainee *result=this;
  while (true)
    {
      if (!result->GetNext()) return result; 
      result=result->GetNext();
    }
}

// supprime de la liste chaine
void ListeChainee::EradicateFromLC(void)
{
  if (!this) return;
  if (Next)
    Next->Prev=Prev;
  if (Prev)
    Prev->Next=Next;        
  Next=NULL;
  Prev=NULL;
}

// renvoit le nombre d'lments de la liste chaine
unsigned long ListeChainee::GetCount()
{
  if (!this) return 0;
  ListeChainee *Current=GetFirst();
  unsigned int result=1;
  while (true)
    {
      if (!Current->GetNext()) return result;
      Current=Current->GetNext();
      result++;
    }
}

// renvoit le nime lment de la liste (on commence  0)
ListeChainee *ListeChainee::GetElementByPosition(unsigned int pos)
{
  if (!this) return NULL;
  ListeChainee *result=GetFirst();
  unsigned int CurrentPos=0;
  while (true)
    {
      if (CurrentPos==pos) return result;
      if (!result->GetNext()) return NULL;
      result=result->GetNext();
      CurrentPos++;
    }
}


int ListeChainee::GetPosition()
{
  if (!this) return -1;
  ListeChainee *Current=GetFirst()->GetNext();
  unsigned long Pos=0;
  while (this!=Current)
    {
      Pos++;
      Current=Current->GetNext();
    }
  return Pos;
}

// rajoute un lment  la fin de la liste
void ListeChainee::AddElement(ListeChainee *ToAdd)
{
  if (!this) return;
  if (!ToAdd) return;
  ListeChainee *last=GetLast();
  if (!last) return;
  last->InsertElementAfter(ToAdd);
}

// fonctions d'insertion d'lments
void ListeChainee::InsertElementBefore(ListeChainee *ToAdd)
{
  if (!this) return;
  if (!ToAdd) return;
  ToAdd->SetPrev(Prev);
  ToAdd->SetNext(this);
  if (Prev) Prev->Next=ToAdd;
  Prev=ToAdd;
}


void ListeChainee::InsertElementAfter(ListeChainee *ToAdd)
{
  if (!this) return;
  if (!ToAdd) return;
  ToAdd->Next=Next;
  ToAdd->Prev=this;
  if (Next) Next->Prev=ToAdd;
  Next=ToAdd;
}


// Code de la classe SortClass utilise dans GetFirstFreeMember dans LC.hpp
