/* lc.hpp
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: lc.hpp,v 1.3 2006/04/10 21:04:16 cvs Exp $
 *
 */
/******************************************************************
 *                      Generic LC library                        *
 *****************************************************************/

// by Bluffy
#ifndef LCHPP
#define LCHPP

#include "lc.h"
#define SORTCLASS_H
//#include "xmem.h"
#ifndef max // on va pas inclure windows.h pour ca...
  #define max(a, b)  (((a) > (b)) ? (a) : (b))
#endif
#ifndef min
  #define min(a, b)  (((a) < (b)) ? (a) : (b))
#endif


// supprime tous les lments de la liste chaine, sauf le Root.
template <class LC> 
void EradicateLC(LC *Root)
{
  if (!Root) return;
  LC *Current=(LC*)(Root->GetLast());
  LC *ToDelete=NULL;
  while (Current)
    {
      Current=(LC*)(Current->GetPrev());
      if (Current)
	{
	  ToDelete=(LC*)(Current->GetNext());
	  ToDelete->EradicateFromLC();
	  delete ToDelete;
	}
    }
}

//idem que precedemment avec CircularLC
template <class LC>
void EradicateCircularLC(LC *Root)
{
  if (!Root) return;
  LC *Current=(LC*)(Root->GetFirst());
  LC *End=(LC*)(Root->GetNextRoot());
  if (!Current) return;
  while (Current!=End)
    {
      Current=(LC*)(Current->GetNext());
      delete (LC*)(Current->GetPrev());
    }
}

/*
 * Renvoit l'lment dont le membre correspond  la valeur fournie.
 *  mais laiisse la lc faire le test avec vool compare
 */
template <class LC>
LC *GetElementByMemberb(LC *current,void * toSearchFor,int MemberCode)
{
  if (!current) return NULL;
  LC *result=current;
  while (true)
    {
      if (!result) return NULL;
      if (result->Compare(toSearchFor,MemberCode)) return result;
      result=(LC*)(result->GetNext());
    }
}

/*
 * GetElementByMember
 *
 * Renvoit l'lment dont le membre correspond  la valeur fournie.
 * cette fonction postule l'existence dans la classe LC de la fonction
 * GetMember(int) et suppose que la classe Type dispose de l'operateur ==
 */
template <class LC,class Type>
LC *GetElementByMember(LC *current,Type toSearchFor,int MemberCode)
{
  if (!current) return NULL;
  Type ToCompare;
  LC *result=(LC*)(current->GetFirst());
  while (true)
    {
      if (!result) return NULL;
      ToCompare=(Type)(result->GetMember(MemberCode));
      if (ToCompare==toSearchFor) return result;
      result=(LC*)(result->GetNext());
    }
}


/*
 * Idem que prcdent mais renvoit le nieme lment verifiant la propriete
 * memes postulats sur la classe LC
 */
template <class LC,class Type>
LC *GetElementByMemberEx(LC *current,Type toSearchFor,int Occurence,int MemberCode)
{
  int OccurencesFound=0;
  if (!current) return NULL;
  LC *result=(LC*)(current->GetFirst());
  while (true)
    {
      if (!result) return NULL;
      if ((Type)(result->GetMember(MemberCode))==toSearchFor)
	{
	  if (++OccurencesFound==Occurence)
	    return result;
	}
      result=(LC*)(result->GetNext());
    }
}


#ifndef SORTCLASS_H
#define SORTCLASS_H


class SortClass : public ListeChainee
{
public :
  SortClass(unsigned long yourMember);
  unsigned long Member;
  unsigned long GetMember(int MemberCode);
  void ReceiveMessage(int MessageCode);
};


/*
 * GetFirstFreeMember
 *
 * renvoit la premiere valeur du membre spcifi qui n'est pas utilise
 * dans la liste chaine.
 */
template <class LC,class Type>
Type GetFirstFreeMember(LC *current,Type From,int MemberCode)
{
  if (!current) return (Type)NULL;
  LC *Courant=(LC*)(current->GetFirst()->GetNext());
  SortClass *Root=new SortClass(0);
  while (Courant)
    {
      SortClass *ToAdd=new SortClass((unsigned long)(Courant->GetMember(MemberCode)));
      InsertByMember<SortClass,unsigned long>(Root,ToAdd,0);
      Courant=(LC*)(Courant->GetNext());
    }
  SortClass *SortCurrent=(SortClass*)(Root->GetNext());
  if ((SortCurrent)&&(SortCurrent->Member>From)) return From;
  while (SortCurrent)
    {
      if (!SortCurrent->GetNext())
	{
	  Type Result=(Type)(max(From,SortCurrent->Member+1));
	  SendMessageToAll<SortClass,int>(Root,0);
	  delete Root;
	  return Result;
	}
      if ((((SortClass*)(SortCurrent->GetNext()))->Member-SortCurrent->Member>1)&&(SortCurrent->Member>=From))
	{
	  Type Result=(Type)(SortCurrent->Member)+1;
	  SendMessageToAll<SortClass,int>(Root,0);
	  delete Root;
	  return Result;
	}
      SortCurrent=(SortClass*)(SortCurrent->GetNext());
    }
  return From; //si sa liste chaine elle est ridicule, l'enervons pas renvoyons lui ce qu'il veut...
}
#endif


/*
 * Envoie le meme message  tous les lments de la liste chaine
 * par l'intermediaire de la fonction postule ReceiveMessage
 */
template <class LC,class Type>
void SendMessageToAll(LC *current,Type message)
{
  if (!current)return;
  LC *toDelete=(LC*)(current->GetLast());
  if (!toDelete) return;
  while (true)
    {
      if (!toDelete->GetPrev()) break;
      toDelete=(LC*)(toDelete->GetPrev());
      ((LC*)(toDelete->GetNext()))->ReceiveMessage(message);
    }
  toDelete->ReceiveMessage(message);
}


/*
 * Inverse deux lments, LA fonction lourde, si qqn a une meilleure
 * implmentation qu'il m'appelle
 */
template <class LC>
void SwapElements(LC *first,LC *second)
{
  // deux cas : ou ils sont colls ou ils ne le sont pas.
  if ((first->GetNext()==second)||(first->GetPrev()==second))
    {
      // colls
      LC* un=NULL;
      LC* deux=NULL;
      if (first->GetNext()==second)
	{
	  un=first;
	  deux=second;
	}
      else
	{
	  un=second;
	  deux=first;
	}
      LC *temp1=(LC*)(un->GetPrev());
      LC *temp2=(LC*)(deux->GetNext());
      if (temp1) temp1->SetNext(deux);
      if (temp2) temp2->SetPrior(un);
      un->SetPrior(deux);
      deux->SetNext(un);
      un->SetNext(temp2);
      deux->SetPrior(temp1);
    }
  else
    {
      // pas colls
      LC *prevUn=(LC*)(first->GetPrev());
      LC *nextUn=(LC*)(first->GetNext());
      LC *prevDeux=(LC*)(second->GetPrev());
      LC *nextDeux=(LC*)(second->GetNext());

      if (prevUn) prevUn->SetNext(second);
      if (nextUn) nextUn->SetPrior(second);
      if (prevDeux) prevDeux->SetNext(first);
      if (nextDeux) nextDeux->SetPrior(first);
      first->SetPrior(prevDeux);
      first->SetNext(nextDeux);
      second->SetPrior(prevUn);
      second->SetNext(nextUn);
    }
}


template <class LC,class Type>
LC * EradicateByMember(LC * Root,Type val,int MemberCode)
{
  LC * nRoot=NULL;
  LC *t,*tn;
  for (t=Root,tn=Root?(LC *)Root->GetNext():0;t;t=tn,tn=tn?(LC *)tn->GetNext():0)
    {
      if (t->GetMember(MemberCode)==val) delete t;
      else
	{
	  if (!nRoot) nRoot=t;
	}
    }
  return nRoot;
}

/*
 * InsertByMember
 *
 * Insert l'lment en fonction de son membre dans la liste chaine
 * l'insertion se fait dans l'ordre croissant, si vous voulez dans
 * l'ordre dcroissant, allez vous faire foutre, demerdez vous, vous
 * avez qu' faire return max-member dans GetMember de votre classe
 */
template <class LC,class Type>
void InsertByMember(LC *current,LC *ToAdd,int MemberCode)
{
  if ((!current)||(!ToAdd)) return;
  LC *result=(LC*)(current->GetFirst());
  if (!result) return;
  Type ToFind=(Type)(ToAdd->GetMember(MemberCode));
  while (true)
    {
      if ((Type)(result->GetMember(MemberCode))>ToFind)
	{
	  result->InsertElementBefore(ToAdd);
	  return;
	}
      if ((LC*)(result->GetNext())==(LC*)NULL)
	{
	  result->InsertElementAfter(ToAdd);
	  return;
	}
      result=(LC*)(result->GetNext());
    }
}

// le nom est bien clair de cette fonction nan ?
template <class LC,class Type>
LC *GetFirstElementAboveMember(LC *current,Type toSearchFor,int MemberCode)
{
  if (!current) return (LC*)NULL;
  LC *result=(LC*)(current->GetFirst());
  while (true)
    {
      if (result==(LC*)NULL) return (LC*)NULL;
      if ((Type)(result->GetMember(MemberCode))>=toSearchFor)
	{
	  return result;
	}
      result=(LC*)(result->GetNext());
    }
}

/*
 * SortByMember
 *
 * Fonction de tri par ordre croissant, c'est sympa mais tres lent, vaut mieux utiliser
 * en permanence InsertByMember. (algo pourri, a chier, pas le temps)
 * A refaire...
 */
template <class LC,class Type,class MemberType>
void SortByMember(LC *current,MemberType MemberCode)
{
  if (!current) return;
  LC *result=NULL;
  bool finished=false;
  while (!finished)
    {
      result=(LC*)(current->GetFirst());
      finished=true;
      while (true)
	{
	  if (result==(LC*)NULL) break;
	  if (result->GetNext())
	    {
	      if ((Type)(((LC*)(result->GetNext()))->GetMember(MemberCode))<(Type)(result->GetMember(MemberCode)))
		{
		  finished=false;
		  SwapElements<LC>(result,(LC*)result->GetNext());
		}
	    }
	  result=(LC*)(result->GetNext());
	}
    }
}



/*************************************************************************
 * Fonctions de conversions tableaux/LC : utiles pour l'archivage des lc.
 * necessitent des fonctions de conversion de memoire : alloc. 
 * remplacer ici alloc par votre fonction  vous :
 *
 * WARNING : objects are not allocated via new but with alloc
 *           so, unless the new operator use also the alloc func to
 *           allocate memory, delete won't work, use then the appropriate
 *           free func.
 *************************************************************************/
#ifdef WIN32
#define alloc(a) LocalAlloc(0,a)
#define free LocalFree
#else
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#define alloc malloc
#define CopyMemory(a,b,c) memcpy(a,b,c)
typedef unsigned long DWORD;
#endif
/*
 * FromArrayToLC
 *
 * Array->LC renvoit le premier des elements d'une liste chainee contenant 
 * tous les elements du tableau
 */
template <class LC>
LC *FromArrayToLC(LC *Array, unsigned long count)
{
  LC *Current=NULL;
  for (int i=0;i<count;i++)
    {
      //LC *New=(LC*)alloc(sizeof(LC));
      LC * New=new LC();
      CopyMemory(New,&Array[i],sizeof(LC));
      Current->SetNext(New);//Pas de panique si current==NULL c prevu. 
      New->SetPrev(Current);
      Current=New;
    }
  Current->SetNext(NULL);
  return (LC*)Current->GetFirst();
}


/*
 * FromLCToArra
 *
 * LC->Array renvoit un tableau dont les elements sont les elements de la 
 * liste chainee dans l'ordre.NOTE : il se trouve que chaque element du 
 * tableau pointe grace  prev et next vers le precedent et suivant donc
 * il s'agit bien d'une liste chainee aussi. Mais si vous utilisez les
 * fonctions de liste chainee apres ce ne sera plus dans l'ordre evidemment,
 * il faudra reappeler FromLCToArray.
 */
template <class LC>
LC *FromLCToArray(LC *Current)
{
  Current=(LC*)(Current->GetFirst());
  LC *Root=Current;
  if (!Root) return NULL;
  LC *Array=(LC*)alloc(Current->GetCount()*sizeof(LC));
  for (int i=0;Current;i++)
    {
      CopyMemory(&Array[i],Current,sizeof(LC));
      Array[i].SetPrev(NULL);
      if (i)
	{
	  Array[i-1].SetNext(&Array[i]);
	  Array[i].SetPrev(&Array[i-1]);
	}
      Current=(LC*)Current->GetNext();
    }
  Array[Root->GetCount()-1].SetNext(NULL);
  return Array;
}


// bearclaw's mod ver:writes at buff and return buff+sz written
template <class LC>
void *FromLCToArray(LC *Current,void * buff)
{
  Current=(LC*)(Current->GetFirst());
  LC *Root=Current;
  if (!Root) return NULL;
  LC *Array=(LC*)buff;//alloc(Current->GetCount()*sizeof(LC));
  int i;
  for (i=0;Current;i++)
    {
      CopyMemory(&Array[i],Current,sizeof(LC));
      Array[i].SetPrev(NULL);
      if (i)
	{
	  Array[i-1].SetNext(&Array[i]);
	  Array[i].SetPrev(&Array[i-1]);
	}
      Current=(LC*)Current->GetNext();
    }
  Array[i].SetNext(NULL);
  return (void *)((unsigned long)buff+(i+1)*sizeof(LC));
}

/************************************************************************
 * Here are two advanced LCToArray and ArrayToLC functions.
 * The difference between them and the previous ones is that for
 * each object of the chained list given or returned, there can be a
 * pointer to a buffer of dynamic size.
 * These functions suppose you have implemented a function in your class:
 * void LCBuffManipulation(DWORD Id,LCBufferStruct *); 
 * This must fill in the LCBufferStruct structure (defined later) if
 * Id==LC_GETBUFF, and set the object pointers if Id=LC_SETBUFF.
 *
 * VERY IMPORTANT NOTE : when this function is called with the LC_GETBUFF
 *   flag, you must not use the Buffer itself to get its size, because
 *   the pointer may not be valid. For example, if this buffer is a mere
 *   char * do not use strlen to get the length, store it in a separate
 *   ember. Even if this pointer is not valid, it is what I want you
 *   to put in the structure, because I will make it valid with an addition.
 *   cf the code of FromBuffToLC for that.
 *
 * Also note all objects/buffers are allocated by the "alloc" function.
 * So you cannot use the delete operator on what I return here.
 ***********************************************************************/

struct LCBufferStruct
{
  void *Pointer;
  unsigned long  Size;
};

#define LC_GETBUFF_BUILD 0
#define LC_GETBUFF_RESTORE 0xbea // rclaw, muhaha
#define LC_SETBUFF 1

/*
 * FromLCToBuff
 *
 * This func return a LCBufferStruct *. The Pointer member of the structure
 * can safely be converted to a LC * pointer, which can be used as an array
 * of LC. Note that each object points to the good buffer.
 * It returns such a structure because Data is stored just at the end of
 * the Array and if you want to store it to a file for example, you may want
 * to know what is the size of the Buffer.
 * The LC in the Array is also a chained list. If you want to know the number
 * of elements, just call GetElementCount, Don't divide Size by sizeof(LC)
 * because there is Data at the end.
*/
template <class LC>
LCBufferStruct *FromLCToBuff(LC *Current)
{
  unsigned long  Count=Current->GetCount();
  LCBufferStruct *StructArray=(LCBufferStruct*)alloc(Count*sizeof(LC));
  Current=(LC*)(Current->GetFirst());
  LC *Trans=Current;
  unsigned long TotalSize=Count*(sizeof(LC)+4)+4; //+4 : cf plus bas
  for (int i=0;Trans;i++)
    {
      Trans->LCBuffManipulation(LC_GETBUFF_BUILD,&StructArray[i]);
      Trans=(LC*)(Trans->GetNext());
      TotalSize+=StructArray[i].Size;
    }
  void *Buffer=alloc(TotalSize);
  // on a calcule la taille totale, on a alloue le buffer, c'est parti.
  Trans=Current;
  // on copie les elements et on les fait pointer tranquilou comme dans LCToArray
  LC *Array=(LC*)Buffer;
  // Le +4 ici pour laisser une marge de 4 entre la fin de la liste chainee
  // et le debut des data
  void *DataPointer=(void*)((unsigned long)Buffer+Count*sizeof(LC)+4);
  for (int i=0;Trans;i++)
    {
      CopyMemory(&Array[i],Trans,sizeof(LC));
      Array[i].SetPrev(NULL);
      if (i)
	{
	  Array[i-1].SetNext(&Array[i]);
	  Array[i].SetPrev(&Array[i-1]);
	}
      CopyMemory(DataPointer,StructArray[i].Pointer,StructArray[i].Size);
      StructArray[i].Pointer=DataPointer;
      Array[i].LCBuffManipulation(LC_SETBUFF,&StructArray[i]);
      // Marge de 4 entre chaque Data, pour menager un 0 au ca ou par exemple
      // c'est des chaines de caracteres en ANSI ou WideChar...
      DataPointer=(void*)((unsigned long)DataPointer+StructArray[i].Size+4);
      Trans=(LC*)Trans->GetNext();
    }
  Array[Count-1].SetNext(NULL);
  // Et voila.
  free(StructArray);
  LCBufferStruct *result=(LCBufferStruct*)alloc(sizeof(LCBufferStruct));  
  result->Pointer=Buffer;
  result->Size=TotalSize;
  return result;
}


/*
 * FromBuffToLC
 * 
 * This function restores a buffer returned by FromLCToBuff to a chained list.
 * Note it ONLY manages to restore buffers returned by FromLCToBuff that have NOT
 * been modified. If you modify an array returned by FromLCToBuff, which is possible,
 * just call again FromLCToBuff before storing itand calling this function.
 */
template <class LC>
LC *FromBuffToLC(LCBufferStruct *Buffer)
{
  void *source=alloc(Buffer->Size);
  CopyMemory(source,Buffer->Pointer,Buffer->Size);
  // c'est parti. L'ide, c'est de postuler que ce qu'il me donne est ce que je lui ai
  // donn dans une vie anterieure, donc que c'est correct et que ca correspond bien 
  // mon format de donne. Je vais simplement remettre tous les pointeurs  jour.
  LC *Array=(LC*)source;
  // c'est ici que j'ai besoin du fait qu'il ait pas touch sa liste chainee :
  // s'il l'a touche, rien ne me dit que le deuxieme element est place en position
  // deux dans le tableau...
  int Delta=(DWORD)(source)+sizeof(LC) - (DWORD)(Array[0].GetNext()); //new=old+(new-old) =>delta = new-old
  LC *Current=Array;
  LCBufferStruct *CurrentStruct=(LCBufferStruct*)alloc(sizeof(LCBufferStruct));   
  while (Current)
    {
      if (Current->GetPrev()) // si c'est null je ferais des catastrophes
	Current->SetPrev((LC*)((DWORD)(Current->GetPrev())+Delta));
      if (Current->GetNext())
	Current->SetNext((LC*)((DWORD)(Current->GetNext())+Delta));
      Current->LCBuffManipulation(LC_GETBUFF_RESTORE,CurrentStruct);
      CurrentStruct->Pointer=(void*)((DWORD)(CurrentStruct->Pointer)+Delta);
      Current->LCBuffManipulation(LC_SETBUFF,CurrentStruct);
      Current=(LC*)(Current->GetNext());
    }
  // et vala
  return Array;
}


#undef alloc
#undef free
#endif // LCHPP
