/* log.C
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: log.C,v 1.3 2003/10/02 19:29:56 cos Exp $
 *
 */

#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <syslog.h>
#include "log.h"

unsigned long LogModes=0;
char Temp[1000];
FILE *f=NULL;
int SYSLOG_MAP[]={LOG_DEBUG,LOG_INFO,LOG_INFO,LOG_NOTICE,LOG_WARNING,LOG_ERR,LOG_CRIT,LOG_CRIT};
char * EventGravityStrings[]=
  {
    "DEBUG",
    "ANNOY",
    "INFO",
    "NOTICE",
    "WARNING",
    "ERROR",
    "FATAL",
    "SPACE-TIME CONTINIUM BREACH"
  };


void LogEvent(EventGravity code,char * format,...)
{ 
  char msg[1024];
  va_list arg;
  va_start(arg,format);
  vsnprintf(msg,1020,format,arg);
  va_end(arg);
  if (LogModes&LOGM_SYSLOG)
    {
      syslog(SYSLOG_MAP[code],msg);
    }
  time_t tme=time(NULL);
  sprintf(Temp,"[%s",ctime(&tme));
  Temp[strlen(Temp)-1]=0;
  if (LogModes&LOGM_NOGRAVITYCODE)
    strcat(&Temp[strlen(Temp)],"]");
  else sprintf(&Temp[strlen(Temp)],"][%s]",EventGravityStrings[code]);
  strcat(Temp,msg);
  strcat(Temp,"\n");
  if (LogModes&LOGM_STDOUT) {
    printf(Temp);
    fflush(stdout);
  }
  if ((LogModes&LOGM_FILE)&&(f)) {fwrite(Temp,1,strlen(Temp),f);fflush(f);}
}


void LogSetup(unsigned long LogFlags,char * progname,char * filename) 
{
  LogModes=LogFlags;
  if (LogModes&LOGM_FILE)
    {
      if (!f) f=fopen(filename,"a");
      if (!f) fprintf(stderr,"Logger, failed to open file '%s' for writting",filename);
    }
  else
    {
      if (f) fclose(f);
      f=NULL;
    }
  if (LogModes&LOGM_SYSLOG)
    {
      openlog(progname,0,LOG_USER);
    }
}
