/* net.h
 *
 * Copyright (C) 2002 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: net.h,v 1.5 2004/11/27 15:10:10 cvs Exp $
 *
 */

#include <stdio.h>
#ifndef __NET_H
#define __NET_H


struct arp_struct {
  unsigned char dst_mac[6];
  unsigned char src_mac[6];
  unsigned char pkt_type[2];  
  unsigned char hw_type[2];
  unsigned char pro_type[2];  
  unsigned char hw_len;    
  unsigned char pro_len;   
  unsigned char arp_op[2];
  unsigned char sender_eth[6];
  unsigned char sender_ip[4]; 
  unsigned char target_eth[6];
  unsigned char target_ip[4]; 
};


struct ip_struct
{
  unsigned char dst_mac[6];
  unsigned char src_mac[6];
  unsigned char pkt_type[2];  
  unsigned char version_header;
  unsigned char services;
  unsigned short total_length;
  unsigned short identification;
  unsigned short flag_fragment_offset;
  unsigned char	ttl;
  unsigned char proto;
  unsigned short header_checksum;
  unsigned char	sender_ip[4];
  unsigned char	target_ip[4];
};


struct udp_struct
{
  unsigned char SrcPort[2];
  unsigned char DstPort[2];
  unsigned char length[2];
  unsigned char checksum[2];

};


struct ether_8023
{
  unsigned char	dst_mac[6];
  unsigned char	src_mac[6];
  unsigned char	length[2];
  unsigned char	dsap;
  unsigned char	ssap;
  unsigned char	control_field;
};


#define NETEXTRACT4(a) (a[0]<<24)+(a[1]<<16)+(a[2]<<8)+a[3]
#define NETEXTRACT2(a) (a[0]<<8)+a[1]
#define NETINSERT2(a,b) _NETINSERT2(a,(unsigned long)b)
#define NETINSERT4(a,b) _NETINSERT4(a,(unsigned long)b)
#define TYPE_ARP 0x0806
#define TYPE_IP  0x0800
#define ARP_REQUEST 1
#define ARP_REPLY   2
#define SAP_NETWARE 0xe0
// if type<MIN_TYPE, frame is IEEE 802.3 and not ethernetII
#define MIN_ETHERII_TYPE 1518


/******************************************************************
 *
 * Inline functions
 *
 *****************************************************************/
inline void _NETINSERT2(unsigned char * dest,unsigned long i)
{
  dest[0]=(i>>8)%256;
  dest[1]=i%256;
}


inline void _NETINSERT4(unsigned char * dest,unsigned long i)
{
  dest[0]=(i>>24)%256;
  dest[1]=(i>>16)%256;
  dest[2]=(i>>8)%256;
  dest[3]=i%256;
}


inline char * IPToStr(unsigned int ip)
{
  static char buff[20];
  sprintf(buff,"%u.%u.%u.%u",ip>>24,(ip>>16)&0xff,(ip>>8)&0xff,ip&0xff);
  return buff;
}


inline unsigned int IsValidIP(char * IP)
{
  unsigned long a,z,e,r;
  return  (sscanf(IP,"%u.%u.%u.%u",&a,&z,&e,&r)==4);
}


inline unsigned long StrToIP(char * IP)
{
  unsigned long a,z,e,r;
  sscanf(IP,"%u.%u.%u.%u",&a,&z,&e,&r);
  return (a<<24)+(z<<16)+(e<<8)+r;
}


inline unsigned long NET_BROADCASTIP(unsigned long ip,unsigned long mask)
{
  // une demonstration par recurence prouve trivialement la justesse de l'algo
  unsigned long baseip=ip&mask;
  unsigned long i=1;
  while ((i&mask)==0) i=i<<1;
  i-=1;
  return baseip+i;
}


inline char * MACToStr(unsigned char * src_mac)
{
  int i;
  static char buff[20];
  sprintf(buff,"%2x:%2x:%2x:%2x:%2x:%2x",src_mac[0],src_mac[1],src_mac[2],src_mac[3],src_mac[4],src_mac[5]);
  for (i=0;buff[i];i++) if (buff[i]==32) buff[i]='0';
  return buff;
}


inline bool StrToMAC(char * StrMAC,unsigned char MAC[6])
{
  unsigned int d1,d2,d3,d4,d5,d6;
  unsigned long cnt=sscanf(StrMAC,"%x:%x:%x:%x:%x:%x",&d1,&d2,&d3,&d4,&d5,&d6);
  if (cnt!=6) return false;
  MAC[0]=d1;
  MAC[1]=d2;
  MAC[2]=d3;
  MAC[3]=d4;
  MAC[4]=d5;
  MAC[5]=d6;
  return true;
}


inline bool IsMaskValid(unsigned long Mask)
{
  unsigned long m=0x80000000;
  while ((m)&&(Mask&m)) (m=m>>1);
  //maintenant y'a pu que des zeros en theorie
  while ((m)&&(!(Mask&m))) m=m>>1;
  return !m;
}

#endif
