/* parseconfig.C - Simple configuration file parser
 *
 * Copyright (C) 2003 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: parseconfig.c,v 1.4 2006/04/10 21:04:16 cvs Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <error.h>
#include <errno.h>
#include <string.h>
#include "parseconfig.h"
void printentry(FILE *out,struct configentry *entries,int index,void *value);

static void trim(char * buffer, int *pos)
{
  while (buffer[*pos]==32) (*pos)++;
}


/*static int feedbuffer(FILE *f,char ** buffer,int *buffersize,int * keepafter)
{
  int startwritepos;
  int readocunt;
  if (*keepafter==0)
    {
      // feed request, but keep evrything=>realloc
      int newsize=(*buffersize*3)/2+1;
      char * newbuffer=(char *)malloc(newsize);
      memcpy(newbuffer,*buffer,*buffersize);
      free(*buffer);
      *buffer=newbuffer;
      startwritepos=*buffersize;
      *buffersize=newsize;
    }
  else
    {
      memmove(*buffer+*keepafter,*buffer,*buffersisze-*keepafter);
      startwritepos=*buffersisze-*keepafter;
      *keepafter=0;
    }
  // fills buffer from startwritepos to *buffersize
  readcount=fread(*buffer+startwritepos,1,*buffersize-startwritepos,file);
  return readcount;
}*/


static int feedline(FILE *f,char ** destbuffer, int * destbuffersize)
{
  int pos=0;
  int escape=0;
  int c;
  int emptyline=1;
  while ((c=fgetc(f))!=EOF)
    {
      if (*destbuffersize<=pos+2)
	{
	  int newsize=(*destbuffersize*3)/2+1;
	  char * newbuffer=(char *)malloc(newsize);
	  if (!newbuffer) return 0;
	  memcpy(newbuffer,*destbuffer,pos);
	  free(*destbuffer);
	  *destbuffer=newbuffer;
	  *destbuffersize=newsize;
	}
      if (escape)
	{
	  escape=0;
	  switch (c)
	    {
	    case '\\':
	      (*destbuffer)[pos++]='\\';
	      break;
	    case 'n':
	      (*destbuffer)[pos++]='\n';
	      break;
	    case '\n':
	      break; // line continuation
	    default:
	      (*destbuffer)[pos++]='\\';
	      (*destbuffer)[pos++]=c;
	      break;
	    }
	  continue;
	}
      else
	{
	  if (c=='\\')
	    {
	      escape=1;
	      emptyline=0;
	      continue;
	    }
	  if (c=='\n')
	    {
	      if (emptyline) 
		{
		  pos=0;
		  continue; // empty line
		}
	      (*destbuffer)[pos++]=0;
	      return pos;
	    }
	  if (c!=32) emptyline=0;
	  (*destbuffer)[pos++]=c;	  
	}
    }
  // a final backslash is ignored 
  return pos;
}

/*
 * parseconfig
 *
 * Parses a configuration file, and fills VALUES according to the array of configentry ENTRIES.
 * returns 0 on success, an error code on failure
 */
unsigned int parseconfig(char * filename, struct configentry *entries, void** values)
{
  FILE *in;
  char * buffer;
  int buffersize=2048; // allocated size
  int pos=0; // current read position in buffer
  int end;
  in = fopen(filename,"r");
  if (in==0) return errno;
  buffer = (char *)malloc(buffersize);
  if (!buffer) return ENOMEM;
  while (feedline(in, &buffer, &buffersize))
    { 
      int length;
      int index;
      pos=0;
      if (buffer[0]=='#') continue; // comment line
      trim (buffer,&pos);
      // try to identify a key
      for (length=0;buffer[length+pos]!=0 && buffer[length+pos]!=32;length++) ;
      // scan keys
      for (index=0; entries[index].key; index++)
	{
	  if ( !strncmp(entries[index].key,buffer+pos,length) && entries[index].key[length]==0 ) // yes it is safe
	    {
	      int epos;
	      struct multiplevalues * mvalues;
	      void * value; // we will put the value here temporarily
	      void ** vvalues; // used to have shorter lines in TYPE_CUSTOM
	      int argcount; // argument count for custom type
	      int argpos;
	      int i; // loop variable
	      pos += length;
	      trim (buffer,&pos);
	      switch (entries[index].type)
		{
		case TYPE_STRING:
		  value = malloc (strlen (buffer+pos) + 1);
		  strcpy (value, buffer+pos );
		  break;
		  
		case TYPE_INTEGER:
		  value = (void *)strtol (buffer+pos, NULL, 0);
		  break;
		  
		case TYPE_BOOLEAN:
		  for (epos=pos; buffer[epos]!=0 && buffer[epos]!=32; epos++) ;
		  buffer[epos]=0; // get rid of extra spaces at end of line
		  if ( !strcasecmp (buffer+pos, "true")
		       || !strcasecmp (buffer+pos, "yes") )
		    {
		      value=(void *)1;
		    }
		  else if ( !strcasecmp (buffer+pos, "false")
			    || !strcasecmp (buffer+pos, "no") ) 
		    {
		      value=(void *)0;
		    }
		  else value=(void *)strtol (buffer+pos, NULL, 0);
		  break;
		  
		case TYPE_CUSTOM:
		  // now this is fun
		  // first of all, we count arguments
		  argcount=0;
		  for (i=0;entries[index].customtype[i];i++)
		    {
		      if (entries[index].customtype[i]=='%') argcount++;
		    }
		  value=malloc(sizeof(void *)*argcount);
		  vvalues=malloc(sizeof(void *)*argcount);
		  
		  // handle type string:
		  // not trivial, to scanf we must dereference the ptr for %s, and 
		  // not dereference it to anything else (clear?)
		  argpos=0;
		  for (i=0;entries[index].customtype[i];i++)
		    {
		      if (entries[index].customtype[i]=='%') 
			{
			  if (entries[index].customtype[i+1]=='s')
			    {
			      ((void **)value)[argpos]=malloc(entries[index].maxstringsize+1);
			      vvalues[argpos]=((void **)value)[argpos];
			    }
			  else
			    vvalues[argpos]= & ((void **)value)[argpos];
			  argpos++;
			}
		    }
		  switch (argcount)
		    {
		    case 1:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0]);
		      break;
		    case 2:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1]);
		      break;
		    case 3:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2]);
		      break;
		    case 4:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3]);
		      break;
		    case 5:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4]);
		      break;
		    case 6:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5]);
		      break;
		    case 7:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6]);
		      break;
		    case 8:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6],vvalues[7]);
		      break;
		    case 9:
		      sscanf(buffer+pos,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6],vvalues[7],vvalues[8]);
		      break;
		    }
		  break;
		  
		}
	      // now handle case of multiple values:
	      if (!entries[index].multiple)
		values[index]=value;
	      else 
		{
		  int required_size;
		  mvalues=(struct multiplevalues *)values[index];
		  if (!values[index])
		    {
		      values[index]=malloc(sizeof(struct multiplevalues));
		      if (!values[index])			
			return ENOMEM;
		      mvalues=(struct multiplevalues *)values[index];
		      mvalues->count=0;
		      mvalues->values=malloc(160);
		      if (!mvalues->values) return ENOMEM;
		      mvalues->allocated=160;
		    }
		  if (entries[index].type!=TYPE_CUSTOM)
		    {
		      required_size= sizeof(void *);
		      argcount=1;
		    }
		  else
		    required_size= sizeof(void *)*argcount;
		  // check available size...
		  if (mvalues->allocated<(mvalues->count+1)* required_size)
		    {
		      int newsize=(mvalues->count+3)* required_size;
		      void * tmp=malloc(newsize);
		      if (!tmp) return ENOMEM;
		      memcpy(tmp,mvalues->values,mvalues->count* required_size);
		      free(mvalues->values);
		      mvalues->values=tmp;
		    }		  
		  // now let's copy value in  mvalues->values
		  if (entries[index].type!=TYPE_CUSTOM)
		    mvalues->values[mvalues->count++]=value;
		  else
		    {
		      memcpy(&mvalues->values[mvalues->count*argcount],value,required_size);
		      mvalues->count++;
		    }
		}
	      continue; // skip the key match check 
	    }
	}
    }
  fclose(in);
  return 0;
}


unsigned int fillconfig(char * filename, struct configentry *entries, void** values)
{
  FILE *out;
  char * buffer;
  int buffersize=2048; // allocated size
  int pos=0; // current read position in buffer
  int end;
  int index;
  int argcount;
  int i;
  out = fopen(filename,"w");
  if (out == 0) return errno;
  buffer = (char *)malloc(buffersize);
  if (!buffer) return ENOMEM;
  for (index=0; entries[index].key; index++)
    {
      if (entries[index].multiple)
	{
	  int required_size;
	  int i;
	  struct multiplevalues * mvalues=(struct multiplevalues *)values[index];
	  if (entries[index].type!=TYPE_CUSTOM)
	    required_size= 1;
	  else
	    {
	      int argcount=0;
	      for (i=0;entries[index].customtype[i];i++)
		{
		  if (entries[index].customtype[i]=='%') argcount++;
		}
	      required_size= argcount;
	    }
	  for (i=0;i<mvalues->count;i++)
	    {
	      printentry(out,entries,index,mvalues->values+required_size*i);
	    }
	}
      else
	{
	  printentry(out,entries,index,values[index]);
	}
      
    }
  fclose(out);
  return 0;
}


void printentry(FILE *out,struct configentry *entries,int index,void *value)
{
  void **vvalues=value;
  int argcount;
  int i;
  fprintf (out, "%s ",entries[index].key);
  switch (entries[index].type)
    {
    case TYPE_STRING:
      fprintf (out, "%s\n",value);
      break;
      
    case TYPE_INTEGER:
      fprintf (out, "%d\n",(long)value);
      break;
      
    case TYPE_BOOLEAN:
      fprintf (out, value?"true\n":"false\n");
      break;
      
    case TYPE_CUSTOM:
      argcount=0;
      for (i=0;entries[index].customtype[i];i++)
	{
	  if (entries[index].customtype[i]=='%') argcount++;
	}
      switch (argcount)
	{
	case 1:
	  fprintf(out,entries[index].customtype,vvalues[0]);
	  break;
	case 2:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1]);
	  break;
	case 3:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2]);
	  break;
	case 4:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3]);
	  break;
	case 5:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4]);
	  break;
	case 6:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5]);
	  break;
	case 7:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6]);
	  break;
	case 8:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6],vvalues[7]);
	  break;
	case 9:
	  fprintf(out,entries[index].customtype,vvalues[0],vvalues[1],vvalues[2],vvalues[3],vvalues[4],vvalues[5],vvalues[6],vvalues[7],vvalues[8]);
	  break;
	  
	}
    default:
      fprintf(out, "\n");
    }
}
