/* arphound.C
 *
 * ArpHound
 *
 * Copyright (C) 2002-2003 Matthieu Nottale <matthieu@nottale.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: arphound.C,v 1.7 2003/10/01 19:48:20 cvs Exp 
 *
 */

/*
 * STRUCTURE :
 *   - un BArbre d'adresses MAC, chacune avec une liste chainee d'IPs
 *   - un tableau de listechainee de MACs, une par IP du subnet, acces instantan
 * ATTENTION : toutes les ips sont stoqus ds le bon sens (a.b.c.d: a poid fort)
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <memory>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/ip.h>
#define Time time
#define BPF_MAJOR_VERSION
#include <pcap.h>
using namespace std;   
#include "../shared/dnsquery.h"
#include "../shared/barbre.h"
#include "../shared/net.h"
#define SORTCLASS_H //do not include some unneccesary templates that may fail with some compilers
#include "../shared/lc.h"
#include "../shared/lc.hpp"
#include "../shared/log.h"
extern "C" {
#include "../shared/parseconfig.h"
}
// ADDED
#include <unistd.h> // geteuid check 
#include "../shared/pidfile.h"
#include <signal.h> // SIG*
#include <sys/stat.h> // umask
// Do we have to run as a daemon?
// END-ADDED


#define DEBUG 0
// maximum number of IP allowed in the subnet to monitor
#define MAX_IPCOUNT 65536



/********************TROUBLE DEFINITIONS**********************/
#define TROUBLE0_COUNT 5
enum trouble_type_0
  {
    // trouble for which only a flag per MAC needs to be stored
    trouble_dhcp_request=0,
    trouble_dhcp_reply,
    trouble_packet_self, // source and destination MAC are the same: switch ARP table fucking attempt
    trouble_packet_in_autoconfigure_network,
    trouble_dhcp_server //DHCP reply for a machine nor registered as a DHCP server 
  };


#define TROUBLE1_COUNT 2
enum trouble_type_1 
  {
    // trouble for which only a couple source MAC/IP needs to be stored 
    trouble_arp_request_for_outside=0, // arp request for an IP outside subnet 
    trouble_arp_reply_for_outside // arp request for an IP outside subnet
  };


#define TROUBLE2_COUNT 6
enum trouble_type_2
  {
    // troubles for which we need to store the source MAC, and a couple MAC/IP
    trouble_arp_reply_source_mismatch=0,  // arp reply: source MAC  does not match replied MAC
    trouble_arp_request_source_mismatch,  // arp request source MAC  does not match request "reply to" MAC
    trouble_arp_request_not_broadcast, // an arp request is not broadcasted
    trouble_arp_reply_broadcast, // an arp reply is broadcasted
    trouble_packet_source_mismatch,  // packet from outside subnet, but src MAC != Gateway MAC
    trouble_packet_destination_mismatch // packet to IP outside subnet, but dst MAC != Gateway MAC
  };


static const char * TROUBLE_STRING_0[]=
  {
    "DHCPREQUEST",
    "DHCPREPLY",
    "PACKETSELF",
    "PACKET_IN_AUTOCONFIGURE_NETWORK",
    "DHCPSERVER"
  };


static const char * TROUBLE_STRING_1[]=
  {
    "ARPREQUEST_OUT",
    "ARPREPLY_OUT"
  };


static const char * TROUBLE_STRING_2[]=
  {
    "ARPREPLY_SOURCE_MISMATCH",
    "ARPREQUEST_SOURCE_MISMATCH",
    "ARPREQUEST_NOT_BROADCAST", //disabled
    "ARPREPLY_BROADCAST",
    "PACKET_SOURCE_MISMATCH",
    "PACKET_DESTINATION_MISMATCH"
  };


/**********************CONFIGURATION**************************/
// configuration parameters and their default values, don't change the order
struct arphound_configure_struct
{
  long allow_override ;
  long run_as_daemon;
  long log_to_file ;
  long log_to_syslog ;
  long log_to_stdout;
  long log_discover ;
  long log_dhcp ;
  long log_dns ;
  long log_outsider;
  unsigned long IPCHANGE_LOG_INTERVAL ;
  unsigned long IPCONFLICT_LOG_INTERVAL ;
  unsigned long TROUBLE_LOG_INTERVAL ;
  char * logfile;
  char * interface;
  char * IP;
  char * mask;
  char * dns_server;
  multiplevalues * routelist;
  multiplevalues * criticalcouplelist;
  multiplevalues * dhcpserverlist;
  multiplevalues * routertosameinterfacelist;
  unsigned long TROUBLE0_LOG_INTERVAL[TROUBLE0_COUNT];
  unsigned long TROUBLE1_LOG_INTERVAL[TROUBLE1_COUNT];
  unsigned long TROUBLE2_LOG_INTERVAL[TROUBLE2_COUNT];
};

struct arphound_configure_struct conf;

/**********************************************************/


// flags value for TIP in Tentry->FirstIP chained list
#define MACIP_SEEN_IN_ARP_REPLY 1
#define MACIP_SEEN_IN_IP_PACKET_SOURCE 2

/**********************************************************/




/**********************STRUCTURE DEFINITIONS*******************/
class TEntry;


class TIP:public ListeChainee // for each MAC entry, a chained list of IPs.
{
public:
  unsigned int IP;
  union {
    time_t LastUsed;
    time_t LastNotify;
  };
  unsigned long flags; //for the troubles: trouble code
  unsigned long count; //number of hit since last log
  TIP() {memset(this,0,sizeof(TIP));}
  unsigned long GetMember(int) {return IP;}  //template callback
};


class MACRef //we have an array of those, 1 per IP. Each one is a chained list , each element points to an entry in the B-Tree.
{
public:
  TEntry * Entry;
  MACRef * Next;
  time_t LastNotify; //this field is only updated for the head of the chained list. We waste 4 bytes per IP conflict (boo boo, faites lui les cornes les enfants).
  MACRef() {memset(this,0,sizeof(MACRef));}
};


class MACIP:public ListeChainee //we have a chained list of those for extra functions: IP outside subnet, IP with no DNS entry.
{
public:
  unsigned int IP;
  char MAC[6];
  time_t LastNotify; 
  unsigned long flags; //trouble code
  unsigned long count; //number of hits since last log
  MACIP() {memset(this,0,sizeof(MACIP));}
};


struct Route //gateway definition
{
  unsigned int RouteIP;
  unsigned int RouteMask;
  unsigned int GatewayIP;
  unsigned char GatewayMAC[6];
};

struct DHCPServer
{
  unsigned char MAC[6];
};


struct CriticalCouple
{
  unsigned int IP;
  unsigned char MAC[6];
};

typedef CriticalCouple  RouterToSameInterface;

class TEntry // the type of the nodes in the B-Tree of MAC addresses
{
public:
  unsigned char MAC[6];        // !!THIS MUST BE THE FIRST FIELD
  TIP * LastIP;                // pointer to most recently used IP in FirstIP chained list
  TIP  FirstIP;                // the IP chained list, first element is always defined, so no need for an extra pointer
  TIP * FirstTroubles1;        // a chained list for all troubles related to this mac
  TIP * LastTroubles1;         // ptr to the tail of the chained list
  unsigned long TroubleCount1; // number of entries in FirstTroubles1 chained list
  MACIP * FirstTroubles2;      // a chained list for each type of trouble
  MACIP * LastTroubles2;       // ptr to the tail of the chained list
  unsigned long TroubleCount2; // number of entries in FirstTroubles2 chained lis
  unsigned long Troubles0;     // troubles that needs only a flag to be identified
  unsigned long Trouble0LastNotify[TROUBLE0_COUNT];
  unsigned long Trouble0Count[TROUBLE0_COUNT]; // number of hit since last log
  time_t LastChange;           // last time LastIP changed in this structure
  time_t LastNotify;           // last time an IP change was written to log about this structure
  unsigned long ChangeCount;   // number of times IP changed since last log
  unsigned long FastestChange; // Smallest time interval between two changes since last log
  TEntry() {memset(this,0,sizeof(TEntry));}
  bool operator ==(TEntry * ent) {return !memcmp(MAC,ent->MAC,6);}
  bool operator >(TEntry * ent) {return memcmp(MAC,ent->MAC,6)>0;}
};



/******************************************************/




/*******************VARIABLES********************/
static const unsigned char BROADCASTMAC[6]={0xff,0xff,0xff,0xff,0xff,0xff};
static const unsigned int AutoconfigureNetworkIP=StrToIP("169.254.0.0");
static const unsigned int AutoconfigureNetworkMask=StrToIP("255.255.0.0");
static const unsigned int MulticastNetworkIP=StrToIP("224.0.0.0");
static const unsigned int MulticastNetworkMask=StrToIP("224.0.0.0");

char Error[255];
pcap_t * Device;    //pcap device
char * DeviceName;
unsigned int CurrentIP;
unsigned int CurrentMask;
unsigned int CurrentBaseIP; //currentip&currentmask, often used
unsigned int ServerIP;      //IP of the DNS server
Route * Routes;
unsigned int RoutesCount;
DHCPServer * DHCPServers;
unsigned int DHCPServersCount; 
CriticalCouple * CriticalCouples;
unsigned int CriticalCouplesCount;
Barbre<TEntry> * Arbre=NULL; //the B-Tree of MAC
MACRef ** IPList=NULL;      //the Array of IPs
MACIP * NoDNS=NULL;      //IP without RDNS entry
CriticalCouple * RouterToSameInterfaceList=NULL;
unsigned int RouterToSameInterfaceCount;
/**************************************************/



/**********************PROTOTYPES********************/
#define IPEXTRACT(a) (a[0]<<24)+(a[1]<<16)+(a[2]<<8)+a[3]
char * IPToStr(unsigned int IP);
char * MACToStr(unsigned char *);
void Sniff(u_char *user, const struct pcap_pkthdr *h, const u_char *buff);
void IPInsert(TEntry * ent,unsigned int SenderIP,unsigned long index);
TEntry * IPChangeDetect(unsigned char *src_mac,unsigned int SenderIP);
void LogIPConflict(unsigned int IP,MACRef * mr,bool notfound);
void LogIPChange(TEntry * Ent,bool notfound,time_t PrevChange,unsigned int  oldlastip);
void LogNoDNS(MACIP *mip);
void LogDHCP(MACIP * mip,unsigned long port);
/****************************************************/


// ADDED
/*
 * go_background
 *
 * Run as a daemon
 * Based on iplogger (C) Mike Edulla + modifications by Topi Miettinen
 * PID file handling based on sysklogd (C) Martin Schulze
 */
void go_background()
{
  // Check PID
  if (check_pid(PID_FILE))
    {
      printf("Already running. Exiting...");
      exit(1);
    }

  if (getppid() != 1)
    {
      signal(SIGTTOU, SIG_IGN);
      signal(SIGTTIN, SIG_IGN);
      signal(SIGTSTP, SIG_IGN);
    }
  daemon(0, 0);
  umask(027);

  // Write PID file
  if (!write_pid(PID_FILE))
    {
      printf("Cannot write pid file. Exiting...\n");
      exit(1);
    }
}

/*
 * die
 *
 * Function executed when the daemon is ended
 */

void die(int sig)
{
  if (conf.run_as_daemon)
    {
      (void) remove_pid(PID_FILE);
    }
  //printf("ArpHound: stopped.");
  exit(0);
}

/*
 * handle_signals
 *
 * Define how to handle the different signals
 */

void handle_signals(void)
{
  signal(SIGTERM, die);
  if (!conf.run_as_daemon)
    {
      signal(SIGINT, die);
    }
  // signal(SIGHUP, sighup);
}
// END-ADDED


/*
 * parse_config
 *
 * Fills configuration variables from a configuration file
 * returns 0 on success, a non-zero value on error
 */
#define STATIC_ENTRIES 20
int parse_config(char * configfile)
{
  struct configentry entries[STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT+TROUBLE2_COUNT+1]= {  //don't change the order, it must match arphound_configure_struct
    {"AllowOverride",TYPE_BOOLEAN,NULL,0,0},
    {"RunAsDaemon",TYPE_BOOLEAN,NULL,0,0},
    {"LogToFile",TYPE_BOOLEAN,NULL,0,0},
    {"LogToSyslog",TYPE_BOOLEAN,NULL,0,0},
    {"LogToStdOut",TYPE_BOOLEAN,NULL,0,0},
    {"LogDiscover", TYPE_BOOLEAN,NULL,0,0},
    {"LogDHCP", TYPE_BOOLEAN,NULL,0,0},
    {"LogDNS", TYPE_BOOLEAN,NULL,0,0},
    {"LogOutsider", TYPE_BOOLEAN,NULL,0,0},
    {"IpChangeLogInterval",TYPE_INTEGER,NULL,0,0},
    {"IpConflictLogInterval",TYPE_INTEGER,NULL,0,0},
    {"TroubleLogInterval",TYPE_INTEGER,NULL,0,0},
    {"LogFileName",TYPE_STRING,NULL,0,0},
    {"NetworkInterface",TYPE_STRING,NULL,0,0},
    {"IP",TYPE_STRING,NULL,0,0},
    {"NetMask",TYPE_STRING,NULL,0,0},
    {"DNSServer",TYPE_STRING,NULL,0,0},
    {"Gateway",TYPE_CUSTOM,"%s %s %s %s",1,20},
    {"CriticalCouple",TYPE_CUSTOM,"%s %s",1,20},
    {"DHCPServer",TYPE_STRING,NULL,1,20},
    {"RouterToSameInterface",TYPE_CUSTOM,"%s %s",1,20},
    {NULL,TYPE_STRING,NULL,0,0}
  };
  // dynamicaly add troubles log intervals
  for (int i=STATIC_ENTRIES;i<STATIC_ENTRIES+TROUBLE0_COUNT;i++)
    {
      entries[i].key=(char *)malloc(strlen(TROUBLE_STRING_0[i-STATIC_ENTRIES])+1+strlen("LogInterval"));
      strcpy(entries[i].key,TROUBLE_STRING_0[i-STATIC_ENTRIES]);
      strcat(entries[i].key,"LogInterval");
      entries[i].type=TYPE_INTEGER;
      entries[i].customtype=NULL;
      entries[i].multiple=0;
      conf.TROUBLE0_LOG_INTERVAL[i-STATIC_ENTRIES]=0;
    }
  for (int i=STATIC_ENTRIES+TROUBLE0_COUNT;i<STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT;i++)
    {
      entries[i].key=(char *)malloc(strlen(TROUBLE_STRING_1[i-(STATIC_ENTRIES+TROUBLE0_COUNT)])+1+strlen("LogInterval"));
      strcpy(entries[i].key,TROUBLE_STRING_1[i-(STATIC_ENTRIES+TROUBLE0_COUNT)]);
      strcat(entries[i].key,"LogInterval");
      entries[i].type=TYPE_INTEGER;
      entries[i].customtype=NULL;
      entries[i].multiple=0;
      conf.TROUBLE1_LOG_INTERVAL[i-(STATIC_ENTRIES+TROUBLE0_COUNT)]=0;
    }
  for (int i=STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT;i<STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT+TROUBLE2_COUNT;i++)
    {
      entries[i].key=(char *)malloc(strlen(TROUBLE_STRING_1[i-(STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT)])+1+strlen("LogInterval"));
      strcpy(entries[i].key,TROUBLE_STRING_1[i-(STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT)]);
      strcat(entries[i].key,"LogInterval");
      entries[i].type=TYPE_INTEGER;
      entries[i].customtype=NULL;
      entries[i].multiple=0;
      conf.TROUBLE2_LOG_INTERVAL[i-(STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT)]=0;
    }
  entries[STATIC_ENTRIES+TROUBLE0_COUNT+TROUBLE1_COUNT+TROUBLE2_COUNT].key=NULL;
  int retval=parseconfig(configfile,entries,(void **)&conf);
  if (retval) return retval;
  // handle route entries
  if (!conf.routelist) 
    {// no route entry defined
      RoutesCount=0;
    }
  else 
    {
      char ** datas=(char **)conf.routelist->values;
      Routes=(Route *)malloc(sizeof(Route)*conf.routelist->count);
      RoutesCount=conf.routelist->count;
      for (int i=0;i<conf.routelist->count;i++)
	{
	  if ( !IsValidIP(datas[4*i])   ||
	       !IsValidIP(datas[4*i+1]) ||
	       !IsValidIP(datas[4*i+2]) ||
	       !StrToMAC(datas[4*i+3],Routes[i].GatewayMAC) )
	    {
	      printf("%d th gateway line is malformed:\n",i);
	      printf("Gateway %s %s %s %s\n",datas[4*i],datas[4*i+1],datas[4*i+2],datas[4*i+3]);
	      return 1;
	    }
	  Routes[i].GatewayIP=StrToIP(datas[4*i+2]);
	  Routes[i].RouteIP=StrToIP(datas[4*i]);
	  Routes[i].RouteMask=StrToIP(datas[4*i+1]);
	  //DEBUG
	  /*
	    char a[20],b[20];
	    strcpy(a,IPToStr( Routes[i].GatewayIP));
	    strcpy(b,IPToStr( Routes[i].RouteIP));
	    printf ("Adding route to %s / %s with gateway %s %s\n",b,IPToStr(Routes[i].RouteMask),a,MACToStr(Routes[i].GatewayMAC));
	  */
	  //ENDDEBUG
	}
    }
  if (!conf.dhcpserverlist)
    {
      DHCPServersCount=0;
    }
  else 
    {
      DHCPServersCount=conf.dhcpserverlist->count;
      DHCPServers=(DHCPServer *)malloc(DHCPServersCount*sizeof(DHCPServer));
      char ** datas=(char **)conf.dhcpserverlist->values;
      for (int i=0;i<DHCPServersCount;i++)
	{
	  if (!StrToMAC(datas[i],DHCPServers[i].MAC))
	    {
	      printf("%d th DHCPServer line malformed\n",i);
	      return 1;
	    }
	}
    }
  if (!conf.routertosameinterfacelist) 
    {
      RouterToSameInterfaceCount = 0;
    }
  else
    {
      RouterToSameInterfaceCount = conf.routertosameinterfacelist->count;
      RouterToSameInterfaceList = new RouterToSameInterface[RouterToSameInterfaceCount];
      char ** datas=(char **)conf.routertosameinterfacelist->values;
      for (int i=0; i<RouterToSameInterfaceCount; i++) {
	if (!IsValidIP(datas[i*2]) || (!StrToMAC(datas[i*2+1],RouterToSameInterfaceList[i].MAC)))
	  {
	    printf ("%d th RouterToSameInterface line malformed:\n",i);
	    printf ("RouterToSameInterface %s %s\n",datas[i*2], datas[i*2+1]);
	    return 1;
	  }
      }
  }
  if (!conf.criticalcouplelist)
    {
      CriticalCouplesCount=0;
    }
  else
    {
      CriticalCouplesCount=conf.criticalcouplelist->count;
      CriticalCouples=(CriticalCouple *)malloc(CriticalCouplesCount*sizeof(CriticalCouple));
      char ** datas=(char **)conf.criticalcouplelist->values;
      for (int i=0;i<CriticalCouplesCount;i++)
	{
	  if (!IsValidIP(datas[i*2]) || (!StrToMAC(datas[i*2+1],CriticalCouples[i].MAC)))
	    {
	      printf ("%d th CriticalCouple line malformed:\n",i);
	      printf ("CriticalCouple %s %s\n",datas[i*2],datas[i*2+1]);
	      return 1;
	    }
	  CriticalCouples[i].IP=StrToIP(datas[i*2]);
	}
    }
  return 0;
}


static const char * usage="Usage: arphound [OPTION] ...\n listen on an interface all incoming packets, and reports various problems.\n"
"  -c config         use the specified configuration file instead of default\n" 
"  -f file           logs also to the specified file (can be given only once overrides file specified in config file)\n"
"  -ns               do not log to syslog\n"
"  -nd               do not run as daemon\n"
"  -ndisc            do not log IP/MAC couples discovery\n"
"  -nout             do not log ARP requests from IP outside subnet\n"
"  -ndns             do not log IP with no RDNS\n"
"  -ch x             minimum interval in seconds between two logs entry when a MAC uses multiples IP\n"
"  -co x             minimum interval in seconds between two logs entry for an IP conflict\n"
"  -tr x             minimum interval in seconds between two logs entry for any trouble\n";


TIP * TIP_Insert(TIP ** root,unsigned int IP, unsigned long flags,bool * added)
{
  for (TIP * tip=*root;tip;tip=(TIP *)tip->GetNext())
    {
      if (tip->IP==IP && tip->flags==flags)
	{
	  *added=false;
	  return tip;
	}
    }
  *added=true;
  TIP * tip=new TIP();
  if (!tip)
    {
      LogEvent(EVENT_FATAL,"out of memory");
      exit(0);
    }
  tip->IP=IP;
  tip->count=1;
  tip->flags=flags;
  if (!*root) *root=tip;
  else (*root)->InsertElementBefore(tip);
  *root=tip;
  tip->LastNotify=0;
  return tip;
}


MACIP * MACIP_Insert(MACIP ** root, unsigned char MAC[6],unsigned int IP,unsigned long flags,bool * added)
{
  for (MACIP * mip=*root;mip;mip=(MACIP *)mip->GetNext())
    {
      if ( (!memcmp(mip->MAC,MAC,6)) && (mip->IP==IP) && mip->flags==flags)
	{
	  // hit
	  mip->count++;
	  *added=false;	 
	  return mip;
	}
    }
  // not found in structure, creating new element
  *added=true;
  MACIP * mip=new MACIP();
  if (!mip)
    {
      LogEvent(EVENT_FATAL,"out of memory");
      exit(1);
    }
  memcpy(mip->MAC,MAC,6);
  mip->count=1;
  mip->IP=IP;
  mip->flags=flags;
  if (!*root) *root=mip;
  else (*root)->InsertElementBefore(mip);
  *root=mip;
  mip->LastNotify=0;
  return mip;
}


inline bool check_critical(unsigned int IP, unsigned char MAC[6])
{
  for (int i=0;i<CriticalCouplesCount;i++)
    {
      if ( (IP==CriticalCouples[i].IP) || !memcmp(MAC,CriticalCouples[i].MAC,6)) return true;
    }
  return false;
}


inline bool check_critical(unsigned int IP)
{
  for (int i=0;i<CriticalCouplesCount;i++)
    {
      if ( IP==CriticalCouples[i].IP ) return true;
    }
  return false;
}


inline bool check_critical(unsigned char MAC[6])
{
  for (int i=0;i<CriticalCouplesCount;i++)
    {
      if (!memcmp(MAC,CriticalCouples[i].MAC,6)) return true;
    }
  return false;
}


void trouble_add_log_2(trouble_type_2 trouble,TEntry *ent,unsigned char MAC[6],unsigned int IP)
{
  bool added;
  char _MAC[20];
  MACIP * mip=MACIP_Insert(&ent->FirstTroubles2,MAC,IP,trouble,&added);
  if (added) ent->TroubleCount2++;
  if (  ( mip->LastNotify==0)  ||  
	( time(NULL)-mip->LastNotify > (conf.TROUBLE2_LOG_INTERVAL[trouble]?conf.TROUBLE2_LOG_INTERVAL[trouble]:conf.TROUBLE_LOG_INTERVAL ) ))
    {
      bool critical=false;
      critical=check_critical(ent->MAC);
      critical=critical|| check_critical(IP,MAC);
      mip->LastNotify=time(NULL);
      strcpy(_MAC,MACToStr(ent->MAC));
      LogEvent(EVENT_NOTICE,"%s; %s%s; %s; %s; %s; %d",TROUBLE_STRING_2[trouble],critical?"!":" ",added?" ":"c",_MAC,MACToStr(MAC),IPToStr(IP),mip->count);
      mip->count=0;
    }
}


void trouble_add_log_1(trouble_type_1 trouble,TEntry *ent,unsigned int IP)
{
  bool added;
 TIP * tip=TIP_Insert(&ent->FirstTroubles1,IP,trouble,&added);
  if (added) ent->TroubleCount1++;
  if (  ( tip->LastNotify==0)  ||  
	( time(NULL)-tip->LastNotify > (conf.TROUBLE1_LOG_INTERVAL[trouble]?conf.TROUBLE1_LOG_INTERVAL[trouble]:conf.TROUBLE_LOG_INTERVAL)))
    {
      bool critical=false;
      critical=check_critical(IP,ent->MAC);
      tip->LastNotify=time(NULL);
      LogEvent(EVENT_NOTICE,"%s; %s%s; %s; %s; %d",TROUBLE_STRING_1[trouble],critical?"!":" ",added?" ":"c",MACToStr(ent->MAC),IPToStr(IP),tip->count);
      tip->count=0;
    }
}


void trouble_add_log_0(trouble_type_0 trouble,TEntry *ent)
{
  ent->Troubles0=ent->Troubles0 | (1<<trouble);
  ent->Trouble0Count[trouble]++;
  if ( (ent->Trouble0LastNotify[trouble]==0) || 
       (time(NULL)-ent->Trouble0LastNotify[trouble]>(conf.TROUBLE0_LOG_INTERVAL[trouble]?conf.TROUBLE0_LOG_INTERVAL[trouble]:conf.TROUBLE_LOG_INTERVAL)  ) )
    {     
      bool critical=false;
      critical=check_critical(ent->MAC);
      LogEvent(EVENT_NOTICE,"%s; %s%s; %s; %d",TROUBLE_STRING_0[trouble],critical?"!":" ",ent->Trouble0LastNotify[trouble]==0?" ":"c",MACToStr(ent->MAC),ent->Trouble0Count[trouble]);
      ent->Trouble0LastNotify[trouble]=time(NULL);
      ent->Trouble0Count[trouble]=0;
    }
}

/******************************************************************
 *
 * Main function
 *
 *****************************************************************/
int main( int argc, char * argv[] )
{
  unsigned long logflags;
  char * config_file="/etc/arphound.conf";
  if (geteuid() != 0) {
    printf("You must be root to run this program.\n");
    exit(1);
  }
  handle_signals();
  // looking if config file is specified on command line
  for (int pos=1;pos<argc;pos++)
    {
      struct stat  st;
      if (!strcmp(argv[pos],"-c"))
	{
	  if (pos==argc-1)
	    {
	      printf("missing file path after '-f', exitting\n");
	      return -1;
	    }
	  if ( ! stat(argv[pos+1],&st) )
	    config_file=strdup(argv[pos+1]);
	  else
	    {
	      printf ("specified file %s does not exist, using default /etc/arphound.conf\n",argv[pos+1]);
	    }
	  
	}
    }
  // parsing configuration file
  // setting defaults
  conf.allow_override= 1 ;
  conf.run_as_daemon = 1;
  conf.log_to_file = 0;
  conf.log_to_syslog = 0;
  conf.log_to_stdout = 1;
  conf.log_discover = 1;
  conf.log_dhcp = 1;
  conf.log_dns = 1;
  conf.IPCHANGE_LOG_INTERVAL = 20;
  conf.IPCONFLICT_LOG_INTERVAL = 20;
  conf.TROUBLE_LOG_INTERVAL = 180;
  conf.logfile=NULL;
  conf.interface=NULL;
  conf.IP=NULL;
  conf.mask=NULL;
  conf.dns_server=NULL;
  conf.routelist=NULL;
  conf.criticalcouplelist=NULL;
  if (parse_config(config_file))
    {
      printf("error opening or parsing configuration file, exiting\n");
      exit (2);
    }
 
  if (conf.allow_override)
    { 
      //parsing command line arguments, overrinding configuration file
      for (int extra=1;extra<argc;extra++)
	{
	   if (!strcmp(argv[extra],"-c"))
	    {
		  extra++;
	      continue; //already handled
	    }
	  if ( (!strcmp(argv[extra],"-h")) || (!strcmp(argv[extra],"--help")) )
	    {
	      printf(usage);
	      exit(0);
	    }
	  if (!strcmp(argv[extra],"-ns"))
	    {
	      conf.log_to_syslog=0;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-nd"))
	    {
	      conf.run_as_daemon = 0;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-ndisc"))
	    {
	      conf.log_discover = 0;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-nout"))
	    {
	      conf.log_outsider = 0;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-ndns"))
	    {
	      conf.log_dns = 0;
	      continue;
	    }
	  if (extra+1>=argc)
	    {
	      printf("missing argument after %s\n",argv[extra]);
	      exit(0);
	    }
	  if (!strcmp(argv[extra],"-ch"))
	    {
	      unsigned long v=strtol(argv[extra+1],NULL,0);
	      conf.IPCHANGE_LOG_INTERVAL=v;
	      extra++;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-co"))
	    {
	      unsigned long v=strtol(argv[extra+1],NULL,0);
	      conf.IPCONFLICT_LOG_INTERVAL=v;
	      extra++;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-tr"))
	    {
	      unsigned long v=strtol(argv[extra+1],NULL,0);
	      conf.TROUBLE_LOG_INTERVAL=v;
	      extra++;
	      continue;
	    }
	  if (!strcmp(argv[extra],"-f"))
	    {
	      conf.log_to_file=1;
	      conf.logfile=argv[extra+1];
	      extra++;
	      continue;
	    }
	  printf("argument %d not understood\n",extra);
	  exit(0);
	}
    }
  // checking if mandatory parameters are set
  if (!conf.interface)
    {
      printf("you must specify an interface, exiting\n");
      exit(1);
    }
  if (!conf.IP || !conf.mask)
    {
      printf("you must specify an IP adress and a netmask, exiting\n");
      exit(1);
    }
  // looking up interface if unspecified (auto)
  if (!strcmp(conf.interface,"auto"))
    {
      DeviceName=pcap_lookupdev(Error);
      if (!DeviceName) 
	{
	  printf("ARPHound: no suitable network device found : %s\n",Error);
	  return -1;
	}
    }
  else DeviceName=conf.interface;
  // printf("ARPHound: utilise le device %s\n",DeviceName);
  /*if (pcap_lookupnet(DeviceName,&CurrentIP,&CurrentMask,Error)<0)
    {
    printf("RPHound: erreur lors de la detection de l'IP: %s\n",Error);
    return -1;
    }*/
  // setting up and initializing log subsystem
  logflags=0;
  if (conf.log_to_file) logflags=logflags|LOGM_FILE;
  if (conf.log_to_stdout) logflags=logflags|LOGM_STDOUT;
  if (conf.log_to_syslog) logflags=logflags|LOGM_SYSLOG;
  logflags=logflags|LOGM_NOGRAVITYCODE;
  LogSetup(logflags,"arphound",conf.logfile);
  CurrentIP=StrToIP(conf.IP);
  CurrentMask=StrToIP(conf.mask);
  CurrentBaseIP=CurrentIP&CurrentMask;
  // CurrentRevBaseIP=(tip[0])+(tip[1]<<8)+(tip[2]<<16)+(tip[3]<<24);
  // calculate number of IPs in subnet.
  int msk;
  int IPCount=1;
  for (msk=1;!(msk&CurrentMask);msk=msk<<1,IPCount=IPCount<<1) {}
  if (IPCount>MAX_IPCOUNT)
    {
      printf("Too many IPs in network, increase MAX_IPCOUNT and recompile.\n");
      return -1;
    }
  // initializing device
  int snaplen=sizeof(ip_struct);
  if (sizeof(arp_struct)>snaplen) snaplen=sizeof(arp_struct);
  Device=pcap_open_live(DeviceName,snaplen, 1, 1000, Error);
  if (!Device)
    {
      printf("ARPHound: error opening device: %s\n",Error);
      return -1;
    }
  // allocating memory for our structures
  IPList=(MACRef **)malloc(sizeof(MACRef *)*IPCount);
  memset(IPList,0,sizeof(MACRef *)*IPCount);
  Arbre=new Barbre<TEntry>(5);
  if (!Arbre)
    {
      printf("ARPHound: error creating B-tree.\n");
      return -1;
    }
  ServerIP=StrToIP(conf.dns_server);
  // inserting  critical entries and route to same interface in tree
  for (int i=0;i<CriticalCouplesCount; i++) {
    TEntry * nent=new TEntry();
    if (!nent)
      {
	LogEvent(EVENT_FATAL,"out of memory");
	exit(0);
      }
    memcpy(nent->MAC,CriticalCouples[i].MAC,6);
    nent->FastestChange=0;
    nent->LastIP=&nent->FirstIP;
    nent->LastChange=time(NULL);
    nent->FirstIP.IP=CriticalCouples[i].IP;
    nent->FirstIP.LastUsed=-1;
    nent->FirstIP.InitLC();
    Arbre->Add(nent);
    Arbre=Arbre->Root();

    MACRef * mr=new MACRef();
    if (!mr)
      {
	LogEvent(EVENT_FATAL,"out of memory");
	exit(0);
      }
    mr->Entry=nent;
    IPList[CriticalCouples[i].IP-CurrentBaseIP]=mr;
  }

 for (int i=0;i<RouterToSameInterfaceCount; i++) {
   TEntry * nent= Arbre->Find ((TEntry *)RouterToSameInterfaceList[i].MAC);
   if (nent)
     continue;   //already inserted as a critical couple
   nent = new TEntry();
   if (!nent)
     {
	LogEvent(EVENT_FATAL,"out of memory");
	exit(0);
      }
    memcpy(nent->MAC,RouterToSameInterfaceList[i].MAC,6);
    nent->FastestChange=0;
    nent->LastIP=&nent->FirstIP;
    nent->LastChange=time(NULL);
    nent->FirstIP.IP=RouterToSameInterfaceList[i].IP;
    nent->FirstIP.LastUsed=-1;
    nent->FirstIP.InitLC();
    Arbre->Add(nent);
    Arbre=Arbre->Root();

    MACRef * mr=new MACRef();
    if (!mr)
      {
	LogEvent(EVENT_FATAL,"out of memory");
	exit(0);
      }
    mr->Entry=nent;
    IPList[RouterToSameInterfaceList[i].IP-CurrentBaseIP]=mr;
  }


  // ADDED - MODIFIED
  if (conf.run_as_daemon) {
    go_background();
  }
  else {
    printf("ARPHound: main loop\n");
  }
  // END-ADDED
  if (pcap_loop(Device,-1,&Sniff,NULL)<0)
    {
      printf("ARPHound: error in sniff loop: %s\n",pcap_geterr(Device));
      return -1;
    }
  printf("ARPHound: exiting\n");
  return 0;
}


void Sniff(u_char *user, const struct pcap_pkthdr *h, const u_char *buff)
{
  if (DEBUG>1) printf("Sniffe:stamp:%u  len:%u\n",h->ts,h->caplen);
  if (h->caplen<sizeof(arp_struct))
    {
      LogEvent(EVENT_WARNING,"[WARN]recieved a packet of size %d",h->caplen);
      return;
    }
  if (!buff)
    {
      LogEvent(EVENT_WARNING,"[WARN]no data present");
      return;
    }
  arp_struct * Packet=(arp_struct *)buff;
  unsigned int SenderIP;
  bool Is_ARP;
  bool Is_DHCP=false;
  // getting source IP
  if (NETEXTRACT2(Packet->pkt_type)==TYPE_ARP)
    {
      SenderIP=IPEXTRACT(Packet->sender_ip);
      Is_ARP=true;
    }
  else
    {
      Is_ARP=false;
      if (NETEXTRACT2(Packet->pkt_type)==TYPE_IP)
	{
	  ip_struct * is=(ip_struct *)buff;
	  SenderIP=IPEXTRACT(is->sender_ip);
	}
      else 
	{
	  if (DEBUG>1) 
	    {
	      printf("unsupported packet type:%x\n",NETEXTRACT2(Packet->pkt_type));	
	      //for (int i=0;i<20;i++)
	      //  {
	      //    printf("%x ",buff[i]);
	      //  }
	      //  printf("\n");
	    }
	  return; // unsupported packet type
	}
    }
  // our index in the IP array 
  unsigned long index=SenderIP-CurrentBaseIP; //warning, cans still be invalid
  // first of all, let's find out TEntry....
  TEntry *ent=NULL;
  if ( ((SenderIP&CurrentMask)!=(CurrentBaseIP)) ||
       (!IPList[index]) ||
       (memcmp(IPList[index]->Entry->MAC,Packet->src_mac,6)))
    {
      // hmmm, not found non-nominal case
      ent=Arbre->Find((TEntry *)Packet->src_mac);
      if (!ent)
	{
	  // never seen this MAC, create the entry
	  ent=new TEntry();
	  memcpy(ent->MAC,Packet->src_mac,6);
	  ent->FastestChange=0;
	  ent->LastIP=&ent->FirstIP;
	  ent->LastChange=time(NULL);
	  if ((SenderIP&CurrentMask)!=(CurrentBaseIP))
	    ent->FirstIP.IP=0; // no ip
	  else
	    {
	      ent->FirstIP.IP=SenderIP;
	      // log discover
	      if (conf.log_discover)
		LogEvent(EVENT_NOTICE,"DISCOVER; %s; %s",IPToStr(SenderIP),MACToStr(Packet->src_mac));
	    }
	  ent->FirstIP.LastUsed=time(NULL);
	  ent->FirstIP.InitLC();
	  Arbre->Add(ent);
	  Arbre=Arbre->Root();
	}
    }
  else ent=IPList[index]->Entry;
  // arp specific part
  if (Is_ARP)
    {
      int op=NETEXTRACT2(Packet->arp_op);
      if (op==ARP_REQUEST)
	{
	  if (memcmp(Packet->dst_mac,BROADCASTMAC,6))
	    {// TROUBLE trouble_arp_request_not_roadcast
	     // TOO COMMON, removed  trouble_add_log_2(trouble_arp_request_not_broadcast,ent,Packet->dst_mac,NETEXTRACT4(Packet->target_ip));
	    }
	  if (memcmp(Packet->src_mac,Packet->sender_eth,6))
	    {// TROUBLE trouble_arp_request_source_mismatch
	      trouble_add_log_2(trouble_arp_request_source_mismatch,ent,Packet->sender_eth,NETEXTRACT4(Packet->target_ip));
	    }
	  if ( (NETEXTRACT4(Packet->target_ip)&CurrentMask)!=(CurrentBaseIP))
	    {// TROUBLE trouble_arp_for_outside
	      trouble_add_log_1(trouble_arp_request_for_outside,ent,NETEXTRACT4(Packet->target_ip));
	    }
	}
      if (op==ARP_REPLY)
	{
	  if (!memcmp(Packet->dst_mac,BROADCASTMAC,6))
	    { // TROUBLE trouble_arp_reply_broadcast
	      trouble_add_log_2(trouble_arp_reply_broadcast,ent,Packet->sender_eth,NETEXTRACT4(Packet->sender_ip));
	    }
	  if (memcmp(Packet->src_mac,Packet->sender_eth,6))
	    {// TROUBLE trouble_arp_reply_source_mismatch
	      trouble_add_log_2(trouble_arp_reply_source_mismatch,ent,Packet->sender_eth,NETEXTRACT4(Packet->sender_ip));
	    }
	  if ( (NETEXTRACT4(Packet->sender_ip)&CurrentMask)!=(CurrentBaseIP))
	    {// TROUBLE trouble_arp_for_outside
	      trouble_add_log_1(trouble_arp_reply_for_outside,ent,NETEXTRACT4(Packet->sender_ip));
	    }
	}
    }
  // generic check
  if (!memcmp(Packet->src_mac,Packet->dst_mac,6))
    { //TROUBLE trouble_packet_self
      trouble_add_log_0(trouble_packet_self,ent);
    }
  
  // dhcp check
  if (NETEXTRACT2 (Packet->pkt_type) == TYPE_IP)
    {
      ip_struct * is=(ip_struct *)buff;
      if (is->proto==IPPROTO_UDP)
	{
	  udp_struct * udp=(udp_struct *)((unsigned long)is+sizeof(ip_struct));
	  unsigned long port=NETEXTRACT2(udp->SrcPort);
	  if (port==68)
	    {
	      Is_DHCP=true;
	      if (conf.log_dhcp)
		trouble_add_log_0(trouble_dhcp_request,ent);
	    }
	  if (port==67)
	    {
	      Is_DHCP=true;
	      if (conf.log_dhcp)
		trouble_add_log_0(trouble_dhcp_reply,ent);
	      if (DHCPServersCount)
		{
		  bool found=false;
		  for (int i=0;i<DHCPServersCount;i++)
		    {
		      if (!memcmp(ent->MAC,DHCPServers[i].MAC,6)) found=true;
		    }
		  if (!found)  trouble_add_log_0(trouble_dhcp_server,ent);
		}
	    }	   
	}
    }

  if ((SenderIP&CurrentMask)!=(CurrentBaseIP))  
    {
      // source IP is outside subnet
      if ((SenderIP&AutoconfigureNetworkMask)==AutoconfigureNetworkIP)
	{
	  trouble_add_log_0(trouble_packet_in_autoconfigure_network,ent);
	  return;
	}
      else {
	if ((RoutesCount)&&(!Is_ARP)&&(!Is_DHCP)) // dhcp will have a stupid source address, we already handled those 
	  {
	    for (int i=0;i<RoutesCount;i++)
	      {
		if ( ((SenderIP&Routes[i].RouteMask)==Routes[i].RouteIP) &&
		     !memcmp(Packet->src_mac,Routes[i].GatewayMAC,6) )
		  return; // packet is realy comming from a gateway
	      }
	    // no gateway matches IP/MAC, trouble	  
	    trouble_add_log_2(trouble_packet_source_mismatch,ent,Packet->dst_mac,SenderIP);
	  }
	return; // packet outside subnet
      }
    }
  if (!Is_ARP)
    {
      ip_struct * is=(ip_struct *)buff;
      unsigned int TargetIP=IPEXTRACT(is->target_ip);
      if ((TargetIP&CurrentMask)!=(CurrentBaseIP) && RoutesCount) 
	{
	  // packet is going outside subnet
	  // ensure the packet is for a gateway MAC
	  // check if it is something else, unrelated
	  if ((TargetIP&MulticastNetworkMask)==MulticastNetworkIP)
	    {// multicast packet
	    }
	  else 
	    {
	      if (!memcmp(Packet->dst_mac,BROADCASTMAC,6))
		{
		  // sent to broadcast MAC, what should we do with this?
		  // probbly nothing
		  // this can be a packet sent to 255.255.255.255, or something else
		  // TODO realy harmless? shouldn't we log?
		}
	      else 
		{
		  
		  bool validroute=false;
		  for (int i=0;i<RoutesCount;i++)
		    {
		      if ( ((TargetIP&Routes[i].RouteMask)==Routes[i].RouteIP) &&
			   !memcmp(Packet->dst_mac,Routes[i].GatewayMAC,6) )
			{// packet is realy going to the appropriate gateway
			  validroute=true;
			  break;
			}
		    }
		  if (!validroute)
		    trouble_add_log_2(trouble_packet_destination_mismatch,ent,Packet->dst_mac,TargetIP);
		}	      
	    }
	}
    }
  /* MOVED
  // checking if IP is in subnet
  if ((SenderIP&CurrentMask)!=(CurrentBaseIP)) 
    {
      // an arp?
      if (!Is_ARP) return;
      // trouble : arp request from IP outside subnet
      // looking for the MAC in our structure
      if (conf.log_outsider)
	{
	  for (MACIP * mip=Outsider;mip;mip=(MACIP *)mip->GetNext())
	    {
	      if (!memcmp(mip->MAC,Paquet->src_mac,6))
		{
		  // found
		  if (difftime(Time(NULL),mip->LastNotify)>conf.OUTSIDER_LOG_INTERVAL)
		    {
		      mip->LastNotify=Time(NULL);
		      char Msg[50];
		      sprintf(Msg,"[OUTARPc]%s;%s",IPToStr(SenderIP),MACToStr(Paquet->src_mac));
		      LogEvent(EVENT_NOTICE,Msg);
		    }
		  return;
		}
	      // not found
	      MACIP * mip=new MACIP();
	      if (!mip)
		{
		  LogEvent(EVENT_FATAL,"{ERR]out of memory");
		  return;
		}
	      memcpy(mip->MAC,Paquet->src_mac,6);
	      mip->LastNotify=Time(NULL);
	      if (!Outsider) Outsider=mip;
	      else Outsider->InsertElementAfter(mip);
	      char Msg[50];
	      sprintf(Msg,"[OUTARP]%s;%s",IPToStr(SenderIP),MACToStr(Paquet->src_mac));
	      LogEvent(EVENT_NOTICE,Msg);
	      return;
	    }
	  return;
	}
    }
  */

  // nominal check: this MAC uses this IP, and no other MAC does the same
  if ((IPList[index])&&(!memcmp(IPList[index]->Entry->MAC,Packet->src_mac,6))&&(!IPList[index]->Next)&&(SenderIP==IPList[index]->Entry->LastIP->IP))
    {
      return; // do nothing
    }
  //ignore MAC of gateways that may route back to the same interface
  for (int i=0; i<RouterToSameInterfaceCount;i++) {
    if (!memcmp(RouterToSameInterfaceList[i].MAC, Packet->src_mac, 6))
      return; 
  }
  // uh? checking collisions in both ways (IP,MAC)
  if (DEBUG>1) printf("..chk1...\n");
  TEntry * nent=IPChangeDetect(Packet->src_mac,SenderIP);
  if (DEBUG>1) printf("..chk2...\n");
  IPInsert(nent,SenderIP,index);
  if (DEBUG>1) printf("..done\n");
}


TEntry * IPChangeDetect(unsigned char *src_mac,unsigned int SenderIP)
{
  // stupid hack to avoid copying the MAC
  TEntry  *ent=(TEntry *)src_mac;
 
  TEntry * present=Arbre->Find(ent);
  if (present)
    {
      if (present->LastIP->IP==0)
	{
	  // special case, the structure was created without any IP
	  present->LastIP->IP=SenderIP;
	  present->LastIP->LastUsed=time(NULL);
	  if (conf.log_discover)
	    LogEvent(EVENT_NOTICE,"DISCOVER; %s; %s",IPToStr(SenderIP),MACToStr(src_mac));
	  return present;
	}
      if (present->LastIP->IP==SenderIP)
	{
	  if (present->LastIP->LastUsed != -1)  //"stay on top" ip
	    present->LastIP->LastUsed=time(NULL);
	  if (DEBUG) printf("Lastused update for %s:%u\n",IPToStr(SenderIP),time(NULL));
	  return present; // all right, this IP was the last one to use this MAC
	}
    
      // another IP last used the MAC. 
      present->ChangeCount++;
      TIP * ip=GetElementByMember<TIP,unsigned int>(&present->FirstIP,SenderIP,0);
      bool notfound=!ip;
      if (notfound)
	{
	  ip=new TIP();
	  if (!ip)
	    {
	      LogEvent(EVENT_FATAL,"[ERR]Out of memory");
	      exit(0);
	    }
	  ip->IP=SenderIP;
	  present->FirstIP.InsertElementAfter(ip);
	}
      unsigned int oldlip=present->LastIP->IP;
      if (ip->LastUsed != -1) {
	ip->LastUsed=time(NULL);
	present->LastIP=ip;
      }
      else oldlip = SenderIP;
      time_t PrevChange=present->LastChange;
      present->LastChange=time(NULL);
      if ((present->LastChange)&& ((!present->FastestChange)||(difftime(present->LastChange,PrevChange)<present->FastestChange))) {present->FastestChange=difftime(present->LastChange,PrevChange);if (!present->FastestChange) present->FastestChange=1;}
      LogIPChange(present,notfound,PrevChange,oldlip);
      return present;
    }
  // new MAC
  TEntry * nent=new TEntry();
  if (!nent)
    {
      LogEvent(EVENT_FATAL,"out of memory");
      exit(0);
    }
  memcpy(nent->MAC,src_mac,6);
  nent->FastestChange=0;
  nent->LastIP=&nent->FirstIP;
  nent->LastChange=time(NULL);
  nent->FirstIP.IP=SenderIP;
  nent->FirstIP.LastUsed=time(NULL);
  nent->FirstIP.InitLC();
  Arbre->Add(nent);
  Arbre=Arbre->Root();
  if (DEBUG)
    {
      // validating insertion
      TEntry * present=Arbre->Find((TEntry *)src_mac);
      if (!present) printf("ARPHound:_error_:inserted MAC not found in B-tree\n");
    }
  if (conf.log_discover)
    LogEvent(EVENT_NOTICE,"DISCOVER; %s; %s",IPToStr(SenderIP),MACToStr(src_mac));
  return nent;
}


void IPInsert(TEntry * ent,unsigned int SenderIP,unsigned long index)
{
  char DNS[255];
  unsigned int IP=ent->LastIP->IP;
  if (!IPList[index])
    {
      // never seen this IP
      MACRef * mr=new MACRef();
      if (!mr)
	{
	  LogEvent(EVENT_FATAL,"out of memory");
	  exit(0);
	}
      mr->Entry=ent;
      IPList[index]=mr;
      if (conf.log_dns)
	{
	  if (DEBUG>1) printf("DNS...\n");
	  unsigned long err=RDNSQuery(IP,ServerIP,DNS);
	  if ((err)&&(err!=ERROR_DNS))
	    printf("ERR; DNS error %u (errno %u)\n",err,errno);
	  if (DEBUG>1) printf(err?"NO\n":"YES\n");
	  if (err)
	    {
	      // not in dns
	      MACIP *mip=new MACIP();
	      if (!mip)
		{
		  LogEvent(EVENT_FATAL,"ERR; out of memory");
		  exit(0);
		}
	      mip->IP=SenderIP;
	      mip->LastNotify=time(NULL);
	      memcpy(mip->MAC,ent->MAC,6);
	      if (!NoDNS) NoDNS=mip;
	      else NoDNS->AddElement(mip);
	      LogNoDNS(mip);
	    }
	}
      return;
    }
  
  // another MAC sits here!
  
  // is our MAC  in the chained list?
  MACRef * mr=IPList[index];
  for (;mr;mr=mr->Next)
    {
      if (!memcmp(ent->MAC,mr->Entry->MAC,6)) break;
    }
  bool notfound=!mr;
  if (!mr)
    {
      // we add ourself on second position
      mr=new MACRef();
      if (!mr)
	{
	  LogEvent(EVENT_FATAL,"out of memory");
	  exit(0);
	}
      mr->Entry=ent;
      mr->Next=IPList[index]->Next;
      IPList[index]->Next=mr;
      if (conf.log_dns)
	{
	  unsigned long err=RDNSQuery(IP,ServerIP,DNS);
	  if (err)
	    {
	      MACIP *mip=new MACIP();
	      if (!mip)
		{
		  LogEvent(EVENT_FATAL,"out of memory");
		  exit(0);
		}
	      mip->IP=SenderIP;
	      mip->LastNotify=time(NULL);
	      memcpy(mip->MAC,ent->MAC,6);
	      if (!NoDNS) NoDNS=mip;
	      else NoDNS->AddElement(mip);
	      LogNoDNS(mip);
	    }
	}
    }
  LogIPConflict(SenderIP,IPList[index],notfound);
  return;
}


void LogIPChange(TEntry * Ent,bool notfound,time_t PrevChange,unsigned int oldlastip)
{
  // check case where last log entry is recent enough
  if ((!notfound)&&(difftime(time(NULL),Ent->LastNotify)<conf.IPCHANGE_LOG_INTERVAL))
    return;
  // building string
  char Log[512];
  Log[0]=0;
  strcat(Log,"IPCHANGE;  ");
  int criticalpos=strlen(Log)-1;
  bool critical=false;
  strcat(Log,notfound?" ; ":"c; ");
  strcat(Log,MACToStr(Ent->MAC));
  sprintf(&Log[strlen(Log)],"; %d; %d; ",Ent->ChangeCount,Ent->FastestChange);
  sprintf(&Log[strlen(Log)],"%s; ",IPToStr(Ent->LastIP->IP));
  critical=critical || check_critical(Ent->LastIP->IP,Ent->MAC);
  sprintf(&Log[strlen(Log)],"%s; ",IPToStr(oldlastip));
  critical=critical || check_critical(oldlastip);
  bool overrun=false;
  for (TIP * ip=&Ent->FirstIP;ip;ip=(TIP*)ip->GetNext())
    {
      if ((ip->IP!=oldlastip)&&(ip->IP!=Ent->LastIP->IP)&&(ip->LastUsed==-1 || difftime(ip->LastUsed,Ent->LastNotify)>0))
	{
	  if (strlen(Log)>=500)
	    {
	      if (!overrun) sprintf(&Log[strlen(Log)], " ...");
	      overrun=true;	      
	    }
	  else sprintf(&Log[strlen(Log)],"%s  ",IPToStr(ip->IP));
	  critical=critical || check_critical(ip->IP);
	}
    }
  //strcat(Log," )");
  //strcat(Log,".\n");
  if (critical) Log[criticalpos]='!';
  LogEvent(EVENT_NOTICE,Log);
  Ent->ChangeCount=0;
  Ent->LastNotify=time(NULL);
}


void LogNoDNS(MACIP *mip)
{
  LogEvent(EVENT_NOTICE,"DNS; %s; %s",IPToStr(mip->IP),MACToStr((unsigned char *)mip->MAC));
}       


void LogIPConflict(unsigned int IP,MACRef * mr,bool notfound)
{
  if (DEBUG) printf("LastNotify: %u\n",mr->LastNotify);
  if ((!notfound)&&(difftime(time(NULL),mr->LastNotify)<conf.IPCONFLICT_LOG_INTERVAL))
    return;
  // is the conflict solved?
  int Active=0;
  for (MACRef *tm=mr;tm;tm=tm->Next)
    {
      TIP * ip=GetElementByMember<TIP,unsigned int>(&tm->Entry->FirstIP,IP,0);
      if (!ip)
	{
	  printf("ARPHound: structure error.\n"); // don't worry, this won't happen
	  continue;
	}
      if ((ip->LastUsed == -1) || difftime(ip->LastUsed,mr->LastNotify)>=0) Active++;
    }
  if ((!notfound)&&(Active<2)) return; // no conflict in the last time period
  char Log[512];
  Log[0]=0;
  bool critical=false;
  strcat(Log,"IPCONFLICT;  ");
  int criticalpos=strlen(Log)-1;
  strcat(Log,notfound?" ; ":"c; ");
  sprintf(&Log[strlen(Log)],"%s; ",IPToStr(IP));
  critical=critical || check_critical(IP);
  bool overrun=false;
  for (MACRef *tm=mr;tm;tm=tm->Next)
    {
      TIP * ip=GetElementByMember<TIP,unsigned int>(&tm->Entry->FirstIP,IP,0);
      if (DEBUG) printf("Lastused cmp:MAC: %s  time:%u  lastnotify:%u  lastused:%u\n",MACToStr(tm->Entry->MAC),time(NULL),mr->LastNotify,ip->LastUsed); 
      if ((ip)&&( (ip->LastUsed == -1) || (difftime(ip->LastUsed,mr->LastNotify)>=0))) 
	{
	  if (strlen(Log)>=500)
	    {
	      if (!overrun) sprintf(&Log[strlen(Log)], " ...");
	      overrun=true;
	    }
	  else sprintf(&Log[strlen(Log)],"%s; ",MACToStr(tm->Entry->MAC));
	  critical=critical || check_critical(tm->Entry->MAC);
	}
      /*
	if ((ip)&&(difftime(ip->LastUsed,mr->LastNotify)>=0)) 
	strcat(Log,"!");
	sprintf(&Log[strlen(Log)],"%s",MACToStr(tm->Entry->MAC));
      */
    }
  mr->LastNotify=time(NULL);//seul le lastnotify du premier (mr), est utilis
  //strcat(Log,"\n");
  if (critical) Log[criticalpos]='!';
  LogEvent(EVENT_NOTICE,Log);
}
