/*  fastarpgen, quick and dirty ARP packet generator
 *  code more or less stolen from arping
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

#ifndef ETH_P_IP
#define ETH_P_IP 0x0800
#endif
extern "C"
{
#if FREEBSD
#include <sys/socket.h>
#include "freebsd.h" 
#endif

#if MACOSX
#include <sys/socket.h>
#include "freebsd.h" 
#endif

#if USE_NETIF
#include <net/if.h>
#include <net/if_arp.h>
#endif

#include <libnet.h>
#include <pcap.h>

#if OPENBSD
#include "openbsd.h"
#endif

#if SOLARIS
#include "solaris.h"
#include "netinet/arp.h"
#endif

}
#include "../shared/net.h"
static const char usage[]= "fastarpgen [OPTIONS] : an ARP packet generator\n"
"  -if xx                  name of the interface to use\n"
"  -d x                    x microseconds delay between each packet\n"
"  -dd x                   x microseconds delay between each IP or MAC change\n"
"  -fm xx:xx:xx:xx:xx:xx   first MAC address to use\n"
"  -lm xx:xx:xx:xx:xx:xx   last MAC address to use\n"
"  -fi xx.xx.xx.xx         first IP to use\n"
"  -li xx.xx.xx.xx         last ip to use\n"
"  -di xx.xx.xx.xx         IP to ARP-request for\n"
"  -m x                    mode:   s=an IP will always be used with the same mac\n"
"                                  i=each IP with evry MAC\n" 
"                                  m=each mac with evry IP\n"
"                                  n=increase IP and MAC together, wrapping will do the shuffle\n"
"  -c x                    number of packets to send for each MAC/IP couple\n"
"  -mc x                   number of MAC loops\n"
"  -ic x                   number of IP loops\n"
"  -h                      displays this message\n\n";
unsigned char FirstMAC[6]={0,0,0,0,0,0};
unsigned char LastMAC[6]={255,255,255,255,255,255};
unsigned long FirstIP=0;
unsigned long LastIP=-1;
unsigned char MAC[6]; //current MAC
unsigned long IP;    //current IP
unsigned long DIP=0;
unsigned long Count=0;
unsigned long IPCount=0;
unsigned long MACCount=0;
unsigned long Delay=1000000;
unsigned long DDelay=0;
unsigned long PacketCount;
time_t begin;
char * ifname=NULL; //interface name
char Mode;

static struct libnet_link_int *linkint;
char ebuf[LIBNET_ERRBUF_SIZE];
static unsigned char *packet;

static unsigned char eth_null[6]={0,0,0,0,0,0};
static unsigned char eth_broadcast[6]={0xff,0xff,0xff,0xff,0xff,0xff};
unsigned long MACLoop=0; //number of loops done
unsigned long IPLoop=0; //number of loops done

void sigint (int sig)
{
  time_t   end=time(NULL);
  unsigned long duration=end-begin;
  printf ("%d packets sent in %d seconds =>%f pps , %f bps\n",PacketCount,duration,((float)PacketCount/(float)duration),
	  ((float)PacketCount*sizeof(arp_struct)*8/(float)duration));
  exit(0);
}
bool parse_args(int argc, char * argv[])
{
  int pos;
  if (argc%2==0)
    {
      printf(usage);
      return false;
    }
  for (pos=1;pos<argc;pos++)
    {
      if (!strcmp(argv[pos],"-h"))
	{
	  printf(usage);
	  return false;
	}
      if (!strcmp(argv[pos],"-d"))
	{
	  Delay=strtol(argv[pos+1],NULL,0);
	  continue;
	}
       if (!strcmp(argv[pos],"-dd"))
	{
	  DDelay=strtol(argv[pos+1],NULL,0);
	  continue;
	}
       if (!strcmp(argv[pos],"-c"))
	{
	  Count=strtol(argv[pos+1],NULL,0);
	  continue;
	}
       if (!strcmp(argv[pos],"-mc"))
	{
	  MACCount=strtol(argv[pos+1],NULL,0);
	  continue;
	}
       if (!strcmp(argv[pos],"-ic"))
	 {
	  IPCount=strtol(argv[pos+1],NULL,0);
	  continue;
	 }
       if (!strcmp(argv[pos],"-m"))
	 {
	   Mode=argv[pos+1][0];
	   if ( Mode!='s' && Mode!='m' && Mode!='i' && Mode!='n' )
	     {
	       printf("invalid mode specified, mode must be s m i or n\n");
	       return false;
	     }
	   continue;
	 }
       if (!strcmp(argv[pos],"-fm"))
	 {
	   int  ok=StrToMAC(argv[pos+1],FirstMAC);
	   if (!ok)
	     {
	       printf ("invalid MAC address format for option -fm\n");
	       return false;
	     }
	  continue;
	 }
       if (!strcmp(argv[pos],"-lm"))
	 {
	   int  ok=StrToMAC(argv[pos+1],LastMAC);
	   if (!ok)
	     {
	       printf ("invalid MAC address format for option -lm\n");
	       return false;
	     }
	  continue;
	 }
       if (!strcmp(argv[pos],"-di"))
	 {
	   int  ok=IsValidIP(argv[pos+1]);
	   if (!ok)
	     {
	       printf ("invalid IP address format for option -di\n");
	       return false;
	     }
	   DIP=StrToIP(argv[pos+1]);
	  continue;
	 }
       if (!strcmp(argv[pos],"-fi"))
	 {
	   int  ok=IsValidIP(argv[pos+1]);
	   if (!ok)
	     {
	       printf ("invalid IP address format for option -fi\n");
	       return false;
	     }
	   FirstIP=StrToIP(argv[pos+1]);
	  continue;
	 }
       if (!strcmp(argv[pos],"-li"))
	 {
	   int  ok=IsValidIP(argv[pos+1]);
	   if (!ok)
	     {
	       printf ("invalid IP address format for option -li\n");
	       return false;
	     }
	   LastIP=StrToIP(argv[pos+1]);
	  continue;
	 }
       if (!strcmp(argv[pos],"-if"))
	 {
	   ifname=strdup(argv[pos+1]);
	   continue;
	 }
    }

  return true;
}

inline int increment_MAC()
{
  if (!memcmp(MAC,LastMAC,6))
    {
      MACLoop++;
      return 1;
    }
  int i=0;
  MAC[0]++;
  for (i=0;MAC[i]==0 && i<5;i++)
    MAC[i+1]++;
  if (!memcmp(MAC,LastMAC,6))
  {
    MACLoop++;
    return 1;
  }
  return 0;
}

inline int increment_IP()
{
  IP++;
  if (IP>=LastIP)
    {
      IP=FirstIP;
      IPLoop++;
      return 1;
    }
  return 0;
}
inline void reset_IP() {IP=FirstIP;}
inline void reset_MAC() { memcpy(MAC,FirstMAC,6);}

void do_increment()
{
  switch (Mode)
    {
    case 's':
      if (increment_IP()) reset_MAC();
      else
	{
	  if (increment_MAC()) reset_IP();
	}
      break;
    case 'i':
      if (increment_MAC()) increment_IP();
      break;
    case 'm':
      if (increment_IP()) increment_MAC();
      break;
    case 'n':
      increment_IP();
      increment_MAC();
      break;
    }
}

int main(int argc, char * argv[])
{
  unsigned char netIP[4];
  unsigned char netDIP[4];
  signal(SIGINT,sigint);
  PacketCount=0;
  time_t end;
  if (!parse_args(argc,argv)) exit(1);
  NETINSERT4(netDIP,DIP);
  IP=FirstIP;
  memcpy(MAC,FirstMAC,6);
  //well, libnet init
  if (!ifname) {
    if (!(ifname = pcap_lookupdev(ebuf))) {
      fprintf(stderr, "pcap_lookupdev(): %s\n", ebuf);
      exit(1);
    }
  }
  if (!(linkint = libnet_open_link_interface(ifname, ebuf))) {
    fprintf(stderr, "libnet_open_link_interface(): %s\n", ebuf);
    exit(1);
  }
  if (-1 == libnet_init_packet(LIBNET_ETH_H + LIBNET_ARP_H
			       + LIBNET_ICMP_H, &packet)) {
    fprintf(stderr, "libnet_init_packet(): error\n");
    exit(1);
  }
  begin=time(NULL);
  while (  (  (!MACCount) || MACLoop<MACCount ) 
	   && (!IPCount) || IPLoop<IPCount)
    {
      NETINSERT4(netIP,IP);
      if (-1 == libnet_build_ethernet(eth_broadcast, // usually xmas
				      MAC, // this box
				      ETHERTYPE_ARP,
				      NULL,
				      0,
				      packet)) 
	{
	  fprintf(stderr, "libnet_build_ethernet(): error\n");
	  exit(1);
	}
      
      if (-1 == libnet_build_arp(ARPHRD_ETHER,
				 ETHERTYPE_IP,
				 6,
				 4,
				 ARPOP_REQUEST,
				 MAC,
				 netIP,
				 eth_null,
				 netDIP,
				 NULL,
				 0,
				 packet + LIBNET_ETH_H)) 
	{
	  fprintf(stderr, "libnet_build_arp(): error\n");
	  exit(1);
	}
      for (int i=0;i<Count;i++)
	{//send on wire
	  if (-1 == (libnet_write_link_layer(linkint,
					     ifname,
					     (u_char*)packet,
					     LIBNET_ARP_H + LIBNET_ETH_H))) {
	    fprintf(stderr, "libnet_write_link_layer(): error\n");
	    exit(1);
	  }
	  PacketCount++;
	  if (Delay) usleep(Delay);
	}
      if (DDelay) usleep(DDelay);
      do_increment();
    } //while

  end=time(NULL);
  unsigned long duration=end-begin;
  printf ("%d packets sent in %d seconds =>%f pps , %f bps\n",PacketCount,duration,((float)PacketCount/(float)duration),
	  ((float)PacketCount*sizeof(arp_struct)*8/(float)duration));
}




